#   Simple Backup - Miscellaneous utilities
#
#   Copyright (c)2008-2010: Jean-Peer Lorenz <peer.loz@gmx.net>
#   Copyright (c)2007-2008: Ouattara Oumar Aziz <wattazoum@gmail.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
"""
:mod:`sbackup.util` -- Miscellaneous utilities
================================================

.. module:: util
   :synopsis: Provide functions for disk operations.
.. moduleauthor:: Ouattara Oumar Aziz (alias wattazoum) <wattazoum@gmail.com>
.. moduleauthor:: Jean-Peer Lorenz <peer.loz@gmx.net>

"""

#TODO: Clean up this module!

from gettext import gettext as _
import os
import subprocess
import tempfile
import inspect
import shutil
import types
import re
import signal

import sbackup
from sbackup.util import local_file_utils
from sbackup.util import exceptions
from sbackup.util import constants
from sbackup.util import structs
from sbackup.util import log


def __get_resource(resource_name, is_file = False):
    """Looks for certain resources installed by sbackup.
    The installation script writes into the 'resources' file where
    the files/resources are being stored.
    This function will look for them and return the appropriate path.
    
    @param resourceName: the ressource name, as complete as possible
    @param isFile: flag whether the resource looked up is a file
    
    @note: The resources file is required to be located in the
            root directory of the sbackup package. 
    """
#    print "Debug: Looking for '%s' (isFile=%s)" % (resourceName, isFile)
    tmp = inspect.getabsfile(sbackup)
    resfile = file(os.path.join(os.path.dirname(tmp), constants.RSRC_FILE), "r")
    resfilelines = resfile.readlines()
    resfile.close()

    for _dir in resfilelines:
        _dir = _dir.strip()
#        print "Debug: Searching in directory '%s'" % _dir
        if os.path.exists(_dir) and os.path.isdir(_dir):
            # only directories stored in resource file are considered 
            if _dir.endswith(resource_name):
                if not is_file:
#                    print "Debug: Directory found in '%s'" % _dir
                    return _dir

            _flist = os.listdir(_dir)
#            print "Debug: directory listing is :" + str(_flist)
            for _item in _flist:
                _path = os.path.join(_dir, resource_name)
                if os.path.exists(_path) and _path.endswith(resource_name):
                    if os.path.isdir(_path):
                        if not is_file:
#                            print "Debug: Directory found in '%s'" % _path
                            return _path
                    else:
                        if is_file:
#                            print "Debug: File found in '%s'" % _path
                            return _path

    raise exceptions.ResourceFileNotFoundError("Unable to find `%s` in the resources list" % resource_name)


def get_resource_file(resource_name):
    return __get_resource(resource_name, is_file = True)


def get_resource_dir(resource_name):
    return __get_resource(resource_name, is_file = False)


def get_locale_dir():
    return get_resource_dir(constants.LOCALE_DIR)


def get_locale_domain():
    return constants.LOCALE_DOMAIN


def get_version_number():
    """Returns the version number that is stored in according 'metainfo' file.
    
    @todo: Implement similar naming as CPython does (version, version_info).
    """
    ver_line = "VERSION=n.a."

    tmp = inspect.getabsfile(sbackup)
    resfile = file(os.sep.join([os.path.dirname(tmp), "metainfo"]), "r")
    resfilelines = resfile.readlines()
    resfile.close()

    for _line in resfilelines:
        _line = _line.strip()
        if _line.startswith("VERSION="):
            ver_line = _line

    versfull_t = ver_line.split("=")
    versfull = versfull_t[1]

    version_t = versfull.split("~")
    vers = version_t[0]
    # version-postfix is currently ignored
    return vers


def launch(cmd, opts, env = None):
    """
    launch a command and gets stdout and stderr
    outStr = a String containing the output from stdout" 
    errStr = a String containing the error from stderr
    retVal = the return code (= 0 means that everything were fine )
    @param cmd: The command to launch
    @return: (outStr, errStr, retVal)
    
    """
    _logger = log.LogFactory.getLogger()

    # Create output log file
    outptr, outFile = tempfile.mkstemp(prefix = "output_")

    # Create error log file
    errptr, errFile = tempfile.mkstemp(prefix = "error_")

    # Call the subprocess using convenience method
    opts.insert(0, cmd)
    _logger.debug("Lauching : " + str(opts))

    retval = subprocess.call(opts, stdin = None, stdout = outptr,
                             stderr = errptr, env = env)

    # Close log handles
    os.close(errptr)
    os.close(outptr)

    outStr = local_file_utils.readfile(outFile)
    errStr = local_file_utils.readfile(errFile)

    local_file_utils.delete(outFile)
    local_file_utils.delete(errFile)

    return (outStr, errStr, retval)


def readlineNULSep(fd, fd1):
    """
    Iterator that read a NUL separeted file as lines 
    @param fd: File descriptor
    @return: the gotten line
    @rtype: String
    """
    _continue = 0

    while _continue < 2 :
        c = fd.read(1)
        currentline = ''

        while c :
            if c == '\0'  :
                # we got a line
                break
            currentline += c
            c = fd.read(1)
        else :
            # c is None
            # This else correspond to the while statement
            _continue += 1

        c1 = fd1.read(1)
        currentline1 = ''

        while c1 :
            if c1 == '\0'  :
                # we got a line
                break
            c1 = fd1.read(1)
            currentline1 += c1
        else :
            # c1 is None
            # This else correspond to the while statement
            _continue += 1

        if _continue == 1 :
            raise exceptions.SBException(\
                                "The length of flist and Fprops are not equals")
        yield (currentline, currentline1)


def is_valid_regexp(aregex):
    """Checks if the given string is a valid regular expression.
    @type aregex: String
    
    @todo: Is an empty expression valid or not? Can we combine both checks?
    """
    if not isinstance(aregex, types.StringTypes):
        raise TypeError("is_valid_regexp: Given parameter must be a string. "\
                        "Got %s instead." % (type(aregex)))
    _res = True
    try:
        dummy = re.compile(aregex)
    except re.error:
        _res = False
    return _res


def is_empty_regexp(aregex):
    """Checks if the given parameter is empty, i.e. is None or a string
    containing only whitespaces.
    
    @type aregex: String 
    """
    if not isinstance(aregex, (types.StringTypes, types.NoneType)):
        raise TypeError("is_empty_regexp: Given parameter must be a string "\
                        "or None. Got %s instead." % (type(aregex)))
    _res = False
    if aregex is None:
        _res = True
    else:
        _stripped_aregex = aregex.strip()
        if _stripped_aregex == "":
            _res = True
    return _res


def add_conf_entry(confline, entry, separator = ","):
    """Appends the given entry to the given configuration line. Entries in
    configurations are separated by specified token.
    
    @param confline:  the string the entry is appended to
    @param entry:      the string that is added
    @param separator: the token that separates the entries
    
    @type confline:      String
    @type entry:      String
    @type separator:  String
    
    @return: the configuration line with the added entry
    @rtype:  String
    
    @raise TypeError: If one of the given parameters is not of string type
    
    @todo: Review behaviour if the entry contains characters equal to the \
            separator.
    """
    __conf_entry_func_type_check(confline, entry, separator)
    _strip_confline = confline.strip(separator)
    if not has_conf_entry(confline, entry, separator):
        _strip_entry = entry
        _line = r"%s%s%s" % (_strip_confline, separator, _strip_entry)
        _line = _line.strip(separator)
    else:
        _line = _strip_confline
    return _line


def remove_conf_entry(confline, entry, separator = ","):
    """Removes the given entry from the given string. Entries in configurations
    are separated by the specified token. Leading and trailing separators are
    taken into account.
    
    @param confline:  the string from which the entry should be removed
    @param entry:      the string that is removed
    @param separator: the token that separates the entries
    
    @type confline:      String
    @type entry:      String
    @type separator:  String
    
    @return: the configuration line without the removed entry
    @rtype:  String
    
    @raise TypeError: If one of the given parameters is not of string type
    """
    __conf_entry_func_type_check(confline, entry, separator)
    _line = r"%s%s%s" % (separator, confline, separator)
    _mentry = r"%s%s%s" % (separator, re.escape(entry), separator)
    _line = re.sub(_mentry , separator, _line)
    _line = _line.strip(separator)
    return _line


def has_conf_entry(confline, entry, separator = ","):
    """Checks whether the given `confline` contains the given
    entry.
    """
    __conf_entry_func_type_check(confline, entry, separator)
    has_entry = False
    conf_t = confline.split(separator)
    for conf_e in conf_t:
        if conf_e == entry:
            has_entry = True
            break
    return has_entry


def __conf_entry_func_type_check(confline, entry, separator):
    """Private helper function that does common type checking
    in the `conf_entry_*` functions.
    """
    if not isinstance(confline, types.StringTypes):
        raise TypeError("Given parameter must be a string. "\
                        "Got %s instead." % (type(confline)))
    if not isinstance(entry, types.StringTypes):
        raise TypeError("Given parameter must be a string. "\
                        "Got %s instead." % (type(entry)))
    if not isinstance(separator, types.StringTypes):
        raise TypeError("Given parameter must be a string. "\
                        "Got %s instead." % (type(separator)))
    return None


def _remove_dups(sequence):
    """Removes duplicate entries from the given list.
    This is not the most efficient implementation, however it
    provides safe behavior.
    """
    if not isinstance(sequence, types.ListType):
        raise TypeError("Expected parameter of type 'list'. Got '%s' instead."\
                        % type(sequence))

    _dest = []
    for val in sequence:
        if val not in _dest:
            _dest.append(val)
    return _dest


def _list_union_no_dups_safe(source_a, source_b):
    """Merges the given lists into a single list not containing any
    duplicate entries using the default way.
    """
    if not isinstance(source_a, types.ListType):
        raise TypeError("Expected parameter of type 'list'. Got '%s' instead."\
                        % type(source_a))
    if not isinstance(source_b, types.ListType):
        raise TypeError("Expected parameter of type 'list'. Got '%s' instead."\
                        % type(source_b))

    _dest = _remove_dups(source_b)
    for val_a in source_a:
        if val_a not in _dest:
            _dest.append(val_a)
    return _dest


def list_union(source_a, source_b):
    """Merges the given lists into a single list not containing any
    duplicate entries in a very efficient way.
    """
    # this functions uses sets in order to merge the lists
    # doing so it is really fast compared to list operations
    # however, this only works if the lists do not contain
    # unhashable entries (such as other lists etc.)
    # in this case the default algorithm is used
    if not isinstance(source_a, types.ListType):
        raise TypeError("Expected parameter of type 'list'. Got '%s' instead."\
                        % type(source_a))
    if not isinstance(source_b, types.ListType):
        raise TypeError("Expected parameter of type 'list'. Got '%s' instead."\
                        % type(source_b))

    _logger = log.LogFactory.getLogger()
    fallback = False
    try:
        _set_a = set(source_a)
        _set_b = set(source_b)
    except TypeError:
        _logger.info("Lists contain unhashable types. Falling back on default.")
        fallback = True

    if fallback:
        _dest = _list_union_no_dups_safe(source_a, source_b)
    else:
        _set_dest = _set_a.union(_set_b)
        _dest = list(_set_dest)

    return _dest


def get_humanreadable_size(size_in_bytes, binary_prefixes = False):
    """Converts given number into more readable values.
     
    @todo: Implement sophisicated class for this!
    @note: Have also a look at function `get_humanreadable_size_str`.
    """
    if isinstance(size_in_bytes, types.StringTypes):
        size_in_bytes = long(size_in_bytes)

    factor = 1000
    if binary_prefixes is True:
        factor = 1024

    _mbytes = size_in_bytes / (factor * factor)
    _kbytes = (size_in_bytes % (factor * factor)) / factor
    _bytes = (size_in_bytes % (factor * factor)) % factor

    return (_mbytes, _kbytes, _bytes)


def get_humanreadable_size_str(size_in_bytes, binary_prefixes = False):
    """Converts given number into readable string.
     
    @todo: Implement sophisicated class for this!
    """
    _mb, _kb, _byt = get_humanreadable_size(size_in_bytes = size_in_bytes, binary_prefixes = binary_prefixes)
    if binary_prefixes is True:
        if _byt == 0:
            _res = _("%(mb)d MiB %(kb)d KiB") % {'mb' : _mb,
                                                 'kb' : _kb}
        else:
            _res = _("%(mb)d MiB %(kb)d KiB %(bytes)d") % {'mb' : _mb,
                                                           'kb' : _kb,
                                                           'bytes' : _byt}
    else:
        if _byt == 0:
            _res = _("%(mb)d MB %(kb)d kB") % {'mb' : _mb,
                                               'kb' : _kb}
        else:
            _res = _("%(mb)d MB %(kb)d kB %(bytes)d") % {'mb' : _mb,
                                                         'kb' : _kb,
                                                         'bytes' : _byt}
    return _res


def enable_timeout_alarm():
    """Helper method that enables timeout alarm handling.
    
    @todo: separate class? should we store the previous signal handler and restore it later? 
    """
    # Set the signal handler
    signal.signal(signal.SIGALRM, sigalarm_handler)


def set_timeout_alarm(timeout):
    """Sets the timeout to the given value.
    """
    signal.alarm(timeout)


def sigalarm_handler(signum, stack_frame): #IGNORE:W0613
    """Signal handler that is connected to the SIGALRM signal.
    
    @raise TimeoutError: A `TimeoutError` exception is raised.
    """
    raise exceptions.TimeoutError("Unable to open device.")


def enable_termsignal():
    signal.signal(signal.SIGTERM, sigterm_handler)


def sigterm_handler(signum, stack_frame): #IGNORE:W0613
    raise exceptions.SigTerminatedError("Application was terminated by signal %s." % signum)


def enable_backup_cancel_signal():
    signal.signal(constants.BACKUP_CANCEL_SIG, sigcancel_handler)


def sigcancel_handler(signum, stack_frame): #IGNORE:W0613
    """
    see `man 7 signal`
    """
    raise exceptions.BackupCanceledError



#   Simple Backup - file access management using FUSE
#
#   Copyright (c)2010,2013: Jean-Peer Lorenz <peer.loz@gmx.net>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
"""
:mod:`sbackup.util.fuse_utils` -- file access management using FUSE
========================================================================

.. module:: fuse_utils
   :synopsis: public entry point to file access management using FUSE
.. moduleauthor:: Jean-Peer Lorenz <peer.loz@gmx.net>

"""

from gettext import gettext as _
import types

from sbackup.util import local_file_utils
from sbackup.util import exceptions
from sbackup.util import interfaces
from sbackup.util import structs
from sbackup.util import pathparse
from sbackup.util import exceptions

#TODO: move available services into FUSE plugin manager and retrieve them dynamically
REMOTE_SERVICE_SFTP = 101
REMOTE_SERVICE_FTP = 102
REMOTE_SERVICE_SSH = 104


REMOTE_SERVICES_AVAIL = { REMOTE_SERVICE_SFTP : _("SFTP"),
                          REMOTE_SERVICE_SSH : _("SSH"),
                          REMOTE_SERVICE_FTP : _("FTP")
                        }
REMOTE_SERVICE_TO_URI_SCHEME = { REMOTE_SERVICE_FTP : pathparse.URI_SCHEME_FTP,
                                 REMOTE_SERVICE_SFTP : pathparse.URI_SCHEME_SFTP,
                                 REMOTE_SERVICE_SSH : pathparse.URI_SCHEME_SSH
                               }
URI_SCHEME_TO_REMOTE_SERVICE = { REMOTE_SERVICE_TO_URI_SCHEME[REMOTE_SERVICE_FTP] : REMOTE_SERVICE_FTP,
                                 REMOTE_SERVICE_TO_URI_SCHEME[REMOTE_SERVICE_SFTP] : REMOTE_SERVICE_SFTP,
                                 REMOTE_SERVICE_TO_URI_SCHEME[REMOTE_SERVICE_SSH] : REMOTE_SERVICE_SSH
                               }


class FuseOperations(interfaces.IOperations):
    """Collects simple operations
    """
    __metaclass__ = structs.Singleton

    pathsep = local_file_utils.PATHSEP

    def __init__(self):
        interfaces.IOperations.__init__(self)
        
    @classmethod
    def chmod_no_rwx_grp_oth(cls, path):
        local_file_utils.chmod_no_rwx_grp_oth(path)

    @classmethod
    def path_exists(cls, path):
        return local_file_utils.path_exists(path)

    @classmethod
    def rename(cls, src, dst):
        local_file_utils.rename(src, dst)

    @classmethod
    def readfile(cls, path):
        return local_file_utils.readfile(path)

    @classmethod
    def normpath(cls, *args):
        return local_file_utils.normpath(*args)

    @classmethod
    def joinpath(cls, *args):
        return pathparse.joinpath(*args)

    @classmethod
    def get_dirname(cls, path):
        return local_file_utils.get_dirname(path)

    @classmethod
    def get_basename(cls, path):
        return local_file_utils.get_basename(path)

    @classmethod
    def openfile_for_write(cls, path):
        return local_file_utils.openfile(path, write = True)

    @classmethod
    def openfile_for_read(cls, path):
        return local_file_utils.openfile(path, write = False)

    @classmethod
    def openfile_for_append(cls, path):
        return local_file_utils.openfile_for_append(path)

    @classmethod
    def pickleload(cls, path):
        return local_file_utils.pickleload(path)

    @classmethod
    def pickledump(cls, datas, path):
        local_file_utils.pickledump(datas, path)

    @classmethod
    def delete(cls, uri):
        local_file_utils.delete(uri)

    @classmethod
    def force_delete(cls, path):
        local_file_utils.force_delete(path)

    @classmethod
    def copyfile(cls, src, dest):
        local_file_utils.copyfile(src, dest)

    @classmethod
    def listdir(cls, path):
        return local_file_utils.listdir(path)

    @classmethod
    def listdir_fullpath(cls, path) :
        return local_file_utils.listdir_fullpath(path)

    @classmethod
    def makedir(cls, target):
        local_file_utils.makedir(target)

    @classmethod
    def makedirs(cls, target) :
        local_file_utils.makedirs(target) #, 0750)

    @classmethod
    def writetofile(cls, path, content) :
        """
        Write a String to a file. You don't have to open and close the file.
        - File = path to the file
        - StringToWrite = String to write into the file.
        """
        local_file_utils.writetofile(path, content)

    @classmethod
    def force_move(cls, src, dst):
        local_file_utils.force_move(src, dst)

    @classmethod
    def is_link(cls, path):
        return local_file_utils.is_link(path)

    @classmethod
    def test_dir_access(cls, path):
        try:
            local_file_utils.listdir(path)
        except Exception, error:
            raise exceptions.FileAccessException(\
                    "Unable to list directory content: %s" % error)

    @classmethod
    def is_dir(cls, path):
        return local_file_utils.is_dir(path)

    @classmethod
    def close_stream(cls, file_desc):
        try:
            file_desc.close()
        except IOError, error:
            raise exceptions.FileAlreadyClosedError(_("Error while closing stream: %s") % error)

def get_scheme_from_service(service):
    if not isinstance(service, types.IntType):
        raise TypeError
    if service not in REMOTE_SERVICES_AVAIL:
        raise ValueError("Given remote service not supported")
    if service not in REMOTE_SERVICE_TO_URI_SCHEME:
        raise ValueError("Given remote service not supported")

    _scheme = REMOTE_SERVICE_TO_URI_SCHEME[service]
    return _scheme


def get_service_from_scheme(scheme):
#    _uri_scheme = self.__uri.uri_scheme
    try:
        _service = URI_SCHEME_TO_REMOTE_SERVICE[scheme]
    except KeyError:
        _service = None
    return _service



#   Simple Backup - snapshot definition
#
#   Copyright (c)2008-2010,2013: Jean-Peer Lorenz <peer.loz@gmx.net>
#   Copyright (c)2007-2008: Ouattara Oumar Aziz <wattazoum@gmail.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
"""
:mod:`sbackup.util.snapshot` -- snapshot definition
=====================================================

.. module:: snapshot
   :synopsis: Defines snapshots
.. moduleauthor:: Jean-Peer Lorenz <peer.loz@gmx.net>
.. moduleauthor:: Ouattara Oumar Aziz <wattazoum@gmail.com>

"""


from gettext import gettext as _
import re
import types


from sbackup.fs_backend import fam
from sbackup.ar_backend import tar

from sbackup.core.ConfigManager import ConfigurationFileHandler

from sbackup.util.exceptions import NotValidSnapshotNameException
from sbackup.util.exceptions import SBException
from sbackup.util.exceptions import NotValidSnapshotException

from sbackup.util import constants
from sbackup.util import structs
from sbackup.util import log


AVAIL_SNP_FORMATS = ["none", "bzip2", "gzip"]


class Snapshot(object):
    """The snapshot class represents one snapshot in the backup directory.
    
    """

    __validname_re = re.compile(r"^(\d{4})-(\d{2})-(\d{2})_(\d{2})[\:\.](\d{2})[\:\.](\d{2})\.\d+\..*?\.(.+)$")


    def __init__ (self, path):
        """The snapshot constructor.
        
        :param path : the path to the snapshot dir.
        
        :todo: Any distinction between creation of a new snapshot and opening\
               an existing one from disk would be useful! The reason is that\
               instantiation a Snapshot with a not existing path creates a\
               snapshot directory in any case, currently! We need to handle the\
               case: opening an snapshot that is supposed to exist but in fact\
               doesn't!
               
        :todo: Separate commit of snapshot and the snapshot (name, path, ...) itself into
               classes with concise responsibilities.

        """
        self.logger = log.LogFactory.getLogger()
        self._fop = fam.get_file_operations_facade_instance()

        # Attributes
        self.__name = False
        self.__base = None
        self.__format = "none" # default value

        self.__followlinks = False

        self.__snarfile = None

        # explicitely defined include and exclude file lists; these lists are filled from the configuration
        self.__includeFlist = structs.SBdict()
        self.__includeFlistFile = None # Str
        self.__excludeFlist = structs.SBdict()
        self.__excludeFlistFile = None # Str

        self.__space_required = constants.SPACE_REQUIRED_UNKNOWN
        self.__splitedSize = 0
        self.__excludes = False

        self.__packages = False
        self.__version = False
        self.__snapshotpath = None

        self.__baseSnapshot = None

        # set some attributes
        self.setPath(path)    # sets path and validates name

        # check if it's an existing snapshot
        if self._fop.path_exists(self.__snapshotpath):
            self.__validateSnapshot(self.__snapshotpath, self.__name)
        else : # Snapshot for creation
            self._fop.makedir(self.__snapshotpath)
            #LP #785512: make the snapshot directory RWX only by owner
            self._fop.chmod_no_rwx_grp_oth(self.__snapshotpath)
    
    def __str__(self):
        "Return the snapshot name"
        return self.getName()

    def set_space_required(self, space):
        if not isinstance(space, (types.IntType, types.LongType)):
            raise TypeError("Parameter of string type expected. Got %s instead." % str(type(space)))
        space = long(space)
        self.__space_required = space

    def get_space_required(self):
        return self.__space_required

    # Public Methods
    def getName(self) :
        " return the name of the snapshot (ie the dir name)"
        if not self.__name :
            raise SBException(_("Snapshot is inconsistent: __name is not set "))
        else :
            return self.__name

    def getDate(self):
        """
        You can get from the dictionary returned the following keys :
        year, month, day, hour, minute, second
        @return : a dictionary with the date time this snapshot has been taken.
        """
        m = re.match(self.__validname_re , self.getName())
        if not m : raise NotValidSnapshotNameException(_("Name of snapshot '%s' doesn't match requirement") % self.getName())
        date = {"year" : int(m.group(1)), "month":int(m.group(2)), "day":int(m.group(3)),
            "hour":int(m.group(4)), "minute":int(m.group(5)), "second":int(m.group(6))}
        return date

    def getIncludeFlist(self):
        """Returns the list of files included into this snapshot.

        @rtype: SBDict
        """
        return self.__includeFlist

    def get_effective_incl_filelist(self):
        """Returns the *effective* list of files included into this snapshot.

        @rtype: SBDict
        """
        return self.__includeFlist.getEffectiveFileList()

    def get_eff_incl_filelst_not_nested(self):
        """Returns the *effective* list of files included into this snapshot.

        @rtype: SBDict
        """
        return self.__includeFlist.get_eff_filelist_not_nested()

    def is_path_in_incl_filelist(self, path):
        """Checks whether the given `path` is contained in list of included files.
        Only full paths (no sub-paths) are considered.
        """
        return self.__includeFlist.hasFile(path)

    def is_subpath_in_incl_filelist(self, path):
        """Checks whether the given `path` is contained in list of included files.
        Full paths as well as sub-paths are considered.
        """
        return self.__includeFlist.contains_path(path)

    def is_path_in_excl_filelist(self, path):
        """Checks whether the given `path` is contained in list of excluded files.
        Only full paths (no sub-paths) are considered.
        """
        return self.__excludeFlist.hasFile(path)

    def disable_path_in_excl_filelist(self, path):
        """Searches for the given `path` in the list of excluded files and set
        the properties to None. Sub-paths are also considered.
        """
        if self.__excludeFlist.has_key(path):
            self.__excludeFlist[path][0] = None

    def disable_path_in_incl_filelist(self, path):
        """Searches for the given `path` in the list of included files and set
        the properties to None. Sub-paths are also considered.
        """
        if self.__includeFlist.has_key(path):
            self.__includeFlist[path][0] = None

    def getExcludeFlist(self):
        """
        get the Exclude file list
        @rtype: list
        """
        return self.__excludeFlist

    def getExcludeFListFile(self):
        """
        @return: the path to the exclude file list file
        """
        if not self.__excludeFlistFile :
            self.__excludeFlistFile = self._fop.joinpath(self.getPath(), "excludes.list")

        return self.__excludeFlistFile

    def getIncludeFListFile(self):
        """
        @return: the path to the include file list file
        """
        if not self.__includeFlistFile :
            self.__includeFlistFile = self._fop.joinpath(self.getPath(), "includes.list")

        return self.__includeFlistFile

    def getSnarFile(self):
        """
        @return: the path to the TAR SNAR file
        """
        if not self.__snarfile :
            self.__snarfile = self._fop.joinpath(self.getPath(), "files.snar")
        return self.__snarfile

    def getPath(self) :
        "return the complete path of the snapshot"
        if not self.__snapshotpath :
            raise SBException(_("Snapshot is inconsistent: __snapshotpath is not set "))
        else :
            return self.__snapshotpath

    def getFormat(self):
        """
        Returns the compression format of the snapshot (from the "format" file or default value)
        """
        _formatf = self._fop.joinpath(self.getPath(), "format")
        if self._fop.path_exists(_formatf):
            self.__format = self._fop.readfile(_formatf).split('\n')[0]
        return self.__format

    def getBase(self) :
        """Returns the name of the base snapshot of this snapshot. If this
        is a full dump, None is returned. Please note that, if a base name
        was successful read from the snapshot directory once, this name is
        returned on any further calls to this method. If you don't want this
        you need to reset `self.__base` to None before.
        
        """
        if not self.__base:
            basefile = self._fop.joinpath(self.__snapshotpath, "base")
            if not self._fop.path_exists(basefile):
                self.__base = None
            else:
                if self.isfull():
                    raise AssertionError("Assertion failed when retrieving "\
                            "snapshot's base: A full backup ('%s') should not "\
                            "have a base file." % self)
                self.__base = self._fop.readfile(basefile).strip()
        return self.__base

    def getBaseSnapshot(self):
        """
        Return the base snapshot (as a Snapshot ) not only the name
        @return: the base Snapshot if it exists or None otherwise (we are a full snapshot) 
        """
        if self.__baseSnapshot is None:
            if not self.isfull():
                if self.getBase():
                    _path = self._fop.get_dirname(self.getPath())
                    _basef = self._fop.joinpath(_path, self.getBase())
                    self.__baseSnapshot = Snapshot(_basef)
        return self.__baseSnapshot

    def getArchive(self):
        """
        Get the snapshot archive which depends on the Format
        @raise NonValidSnapshotException: if the archive equivalent to the described format doesn't exist
        @return: the path to the archive
        """
        problem = False

        if self.getFormat() == "none":
            _arn = self._fop.joinpath(self.getPath(), "files.tar")
            if self._fop.path_exists(_arn):
                return _arn
            else :
                problem = True

        elif self.getFormat() == "gzip":
            _arn = self._fop.joinpath(self.getPath(), "files.tar.gz")
            if self._fop.path_exists(_arn):
                return _arn
            else :
                problem = True

        elif self.getFormat() == "bzip2":
            _arn = self._fop.joinpath(self.getPath(), "files.tar.bz2")
            if self._fop.path_exists(_arn) :
                return _arn
            else :
                problem = True

        else:
            problem = True

        if problem :
            raise NotValidSnapshotException(_("The snapshot compression format is supposed to be '%s' but the corresponding well named file wasn't found") % self.getFormat())

    def getVersion(self) :
        """Retrieves and returns the version of the snapshot.

        """
        if self.__version:
            return self.__version
        elif ":" in self.getName():
            self.__version = "1.0"
            return self.__version
        else:
            verfile = self._fop.joinpath(self.getPath(), "ver")
            if not self._fop.path_exists(verfile):
                return False
            else :
                ver = self._fop.readfile(verfile)
#                self.logger.debug("Version read from snapshot: `%s`" % ver)
                try :
                    # major = 
                    int(ver[0])
                    # minor = 
                    int(ver[2])
                except Exception:
                    raise SBException(_("%(file)s doesn't contain valid value. Ignoring incomplete or non - backup directory. ") % {"file" : verfile})
                self.__version = ver[:3]
                return self.__version

    def getExcludes(self) :
        "Return the content of excludes (the list of Regex excludes)"
        if self.__excludes : return self.__excludes
        else :
            excludefile = self._fop.joinpath(self.getPath(), "excludes")
            if not self._fop.path_exists(excludefile):
                return False
            else :
                self.__excludes = self._fop.pickleload(excludefile)
                return self.__excludes

    def getPackages(self) :
        "Return the packages"
        if self.__packages : return self.__packages
        else :
            packagesfile = self._fop.joinpath(self.getPath(), "packages")
            if not self._fop.path_exists(packagesfile):
                return False
            else :
                self.__packages = self._fop.readfile(packagesfile)
                return self.__packages

    def getSnapshotFileInfos(self, useMem = False, writeFlag = False):
        """Returns a wrapper for the SnapshotFile resp. DirectoryFile that
        contains information what files are stored in this snapshot.

        @param useMem: use or not the memory to store infos about the SNAR file
        @type useMem:  boolean
        @param writeFlag: Will be passed to the SnapshotFile to permit writing
        @type writeFlag:  boolean

        @return: the corresponding SnapshotFile (Mem or Proc)

        @note: The only method usable afterward is getContent(path)
        """
        snpfile = tar.SnapshotFile(self.getSnarFile(), writeFlag)
        snpfileinfo = None

        if useMem:
            snpfileinfo = tar.MemSnapshotFile(snpfile)
        else :
            snpfileinfo = tar.ProcSnapshotFile(snpfile)

        return snpfileinfo

    def getSplitedSize(self):
        """
        @return: the size of each archive in the snapshot (0 means unlimited)

        @todo: Implement CQS pattern !
        """
        _formatf = self._fop.joinpath(self.getPath(), "format")
        if self._fop.path_exists(_formatf):
            self.__splitedSize = int(self._fop.readfile(_formatf).split('\n')[1])
        return self.__splitedSize

    def isfull(self):
        """
        @return: True if the snapshot is full and false if inc
        """
        _name = str(self.getName())
        return _name.endswith(".ful")

    def isFollowLinks(self):
        return self.__followlinks

    def commit(self, targethandler, publish_progress = False, supports_publish = True):
        """Commit snapshot data (i.e. write to disk)
        
        :param publish_progress: Flag whether to call checkpoint script
        :param use_io_pipe: Flag whether to use pipes instead of files defined as parameters for archive
                            writing (should be hidden in FAM)
        
        :todo: Hide implementation detail `use_io_pipe`. 
        """
        if not self.isfull():
            self.commitbasefile()
        self.commitFormatfile()
        self.commitexcludefile()
        self.commitpackagefile()
        self.commitflistFiles()
        self.__commit_archive(targethandler, publish_progress, supports_publish)
        self.commitverfile()

    def addToIncludeFlist (self, item) :
        """
        Add an item to be backup into the snapshot.
        Usage :  addToIncludeFlist(item) where
        - item is the item to be add (file, dir, or link)

        The `include flist` is of type `SBDict`, the according `props` for a single entry
        is '1' for included items.
        """
        self.__includeFlist[item] = "1"

    def addToExcludeFlist (self, item) :
        """
        Add an item to not be backup into the snapshot.
        Usage :  addToExcludeFlist(item) where
        - item is the item to be add (file, dir, or link)

        The `exclude flist` is of type `SBDict`, the according `props` for a single entry
        is '0' for excluded items.
        """
        self.__excludeFlist[item] = "0"

    def check_and_clean_flists(self):
        """Checks include and exclude flists for entries contained in both lists.
        Entries stored in both lists are removed from the exclude list (include overrides
        exclude).

        In theory it is impossible but what's in the case of manually written configuration files?
        
        @todo: Implement this method.
        """
        pass

    # Setters    
    def setFormat(self, cformat = None):
        """
        Sets the backup compression format
        cformat : the format to set
        """
        if cformat in AVAIL_SNP_FORMATS :
            self.logger.debug("Set the compression format to %s" % cformat)
            self.__format = cformat

    def setPath(self, path) :
        "Set the complete path of the snapshot. That path will be used to get the name of the snapshot"
        self.__snapshotpath = self._fop.normpath(str(path))
        name = self._fop.get_basename(self.__snapshotpath)
        if not self.__isValidName(name) :
            raise NotValidSnapshotNameException(_("Name of Snapshot not valid : %s") % self.__name)
        else :
            self.__name = name

    def setBase(self, baseName) :
        """Sets `baseName` as the name of the base snapshot of this
        snapshot and clears the reference to base snapshot object.
        It is not possible to set the base for a full backup, this
        raises a `SBException`. The `baseName` is checked for validity.
        Note that the base of the snapshot is not committed to disk
        if it is set using this method. Call `commitbasefile` for this.
        
        """
        if self.isfull():
            self.__base = None
            self.__baseSnapshot = None
            raise SBException("Base cannot be set for full snapshot.")
        if not self.__isValidName(baseName) :
            raise SBException (_("Name of base not valid : %s") % self.__name)
        # set the name and clean the baseSnapshot
        self.__base = baseName
        self.__baseSnapshot = None

    def setVersion(self, ver = "1.5") :
        "Set the version of the snapshot"
        self.__version = ver

    def setExcludes(self, excludes) :
        "Set the content of excludes (the list of Regex excludes)"
        self.__excludes = excludes

    def setPackages(self, packages = "") :
        """
        set the packages list for debian based distros
        @param packages: Must be the results of the 'dpkg - -get - selections' command . default = '' 
        """
        self.__packages = packages

    def setSplitedSize(self, size):
        """
        @param size: The size in KiB to set
        
        """
        if type(size) != int:
            raise TypeError("The size parameter must be an integer")
        self.__splitedSize = size

    def setFollowLinks(self, activate):
        """
        @param activate: boolean to activate symlinks follow up 
        """
        if type(activate) != bool :
            raise TypeError("the activate parameter must be a boolean")
        self.__followlinks = activate

    # Private
    def __validateSnapshot(self, path, name):
        """
        Validate the snapshot
        @param path : the snapshot path
        @param name : the snapshot name 
        """
        # validate the name
        if not self.__isValidName(self.__name) :
            raise NotValidSnapshotNameException(_("Invalid name of snapshot: %s") % self.__name)

        _verf = self._fop.joinpath(self.getPath(), "ver")
        if  not self._fop.path_exists(_verf):
            raise NotValidSnapshotException(_("Unable to read mandatory file `ver`"))

    def __isValidName(self, name) :
        " Check if the snapshot name is valid "
        _res = False
        if re.match(self.__validname_re , name) is not None:
            if name.endswith(".ful") or name.endswith(".inc"):
                _res = True
        return _res

    def commitFormatfile(self):
        """
        writes the format file
        """
        formatInfos = self.getFormat() + "\n"
        formatInfos += str(self.getSplitedSize())

        _formatf = self._fop.joinpath(self.getPath(), "format")
        self._fop.writetofile(_formatf, formatInfos)

    def commitverfile(self) :
        """Commit ver file on the disk.
        """
        if not self.getVersion():
            self.setVersion()
        _ver = self.getVersion()
        _verf = self._fop.joinpath(self.getPath(), "ver")
        self._fop.writetofile(_verf, _ver)
        self.logger.debug("Commit `ver` file `%s` with info `%s`" % (_verf, _ver))

    def commitbasefile(self):
        """In case this snapshot is an incremental snapshot, base file is
        committed to the disk. If not, this method shouldn't be called.
        The absence of a base for an incremental backup raises a SBException.
        
        """
        if self.isfull():
            self.logger.debug("WARNING: Attempt of committing base file for "\
                              "full snapshot ' % s'." % self.getName())
        else:
            if self.getBase():
                _basef = self._fop.joinpath(self.getPath(), "base")
                self._fop.writetofile(_basef, self.getBase())
            else:
            # base file was not found or base wasn't set. It MUST be full backup
                raise SBException(_("Base name must be set for incremental backup."))

    def commitexcludefile(self):
        """
        Commit exclude file on the disk (the list of Regex excludes).
        @raise SBException: if excludes hasn't been set 
        """
        _exf = self._fop.joinpath(self.getPath(), "excludes")
        self._fop.pickledump(self.__excludes, _exf)

    def commitflistFiles(self):
        """
        Commit the include.list and exclude.list to the disk.
        @todo: simplify the written files! The *.tmp files are partly obsolete.
        """
        if self._fop.path_exists(self.getIncludeFListFile()):
            raise SBException("includes.list should not exist at this stage")
        if self._fop.path_exists(self.getExcludeFListFile()):
            raise SBException("excludes.list should not exist at this stage")

        self._fop.writetofile(self.getIncludeFListFile(),
                              "\n".join(self.__includeFlist.getEffectiveFileList()))
        self._fop.writetofile(self.getExcludeFListFile(),
                              "\n".join(self.__excludeFlist.getEffectiveFileList()))

        # commit temporary lists
#FIXME: unify creation and storage of tmp. files and names
        tmp_incl = self._fop.normpath(ConfigurationFileHandler().get_user_tempdir(),
                                     self._fop.get_basename(self.getIncludeFListFile()))
        tmp_excl = self._fop.normpath(ConfigurationFileHandler().get_user_tempdir(),
                                     self._fop.get_basename(self.getExcludeFListFile()))

        self._fop.writetofile(tmp_incl,
                              "\n".join(self.__includeFlist.get_eff_filelist_not_nested()))
        self._fop.writetofile(tmp_excl,
                              "\n".join(self.__excludeFlist.getEffectiveFileList()))

    def commitpackagefile(self):
        " Commit packages file on the disk"
        _packf = self._fop.joinpath(self.getPath(), "packages")
        if not self.getPackages() :
            self._fop.writetofile(_packf, "")
        else:
            self._fop.writetofile(_packf, self.getPackages())

    def read_excludeflist_from_file(self):
        fe = self._fop.openfile_for_read(self.getExcludeFListFile())
        _res = set(line.strip("\n") for line in fe)
        fe.close()
        return _res

    def __commit_archive(self, targethandler, publish_progress, supports_publish):
        " Make the backup on the disk "
        tar.mk_archive(self, targethandler, publish_progress, supports_publish)



#   Simple Backup - process a certain profile (a distinct configuration)
#
#   Copyright (c)2008-2010,2013: Jean-Peer Lorenz <peer.loz@gmx.net>
#   Copyright (c)2007-2008: Ouattara Oumar Aziz <wattazoum@gmail.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
"""
:mod:`BackupProfileHandler` --- backup handler class
====================================================================

.. module:: BackupProfileHandler
   :synopsis: Defines a backup handler class
.. moduleauthor:: Jean-Peer Lorenz <peer.loz@gmx.net>
.. moduleauthor:: Ouattara Oumar Aziz (alias wattazoum) <wattazoum@gmail.com>

"""


from gettext import gettext as _
import os
import datetime
import socket
import time


from sbackup.fs_backend import fam
from sbackup.core.SnapshotManager import SnapshotManager
from sbackup.core import snapshot

from sbackup import util
from sbackup.util import filecollect
from sbackup.util import exceptions
from sbackup.util import constants
from sbackup.util import log


class BackupProfileHandler(object):
    """Class that handles/manages the backup process of a single profile.
    
    """

    def __init__(self, configmanager, backupstate, dbus_connection = None,
                 use_indicator = False, full_snapshot = False):
        """The `BackupProfileHandler` Constructor.

        :param configmanager : The current configuration manager
        :param backupstate: object that stores the current state of the process
        
        :note: Make sure to call for the appropriate logger before\
               instantiating this class!
               
        """
        self.logger = log.LogFactory.getLogger()
#TODO: Simplify/refactor these attributes.
        self.__dbus_conn = dbus_connection
        self.__use_indicator = use_indicator
        self.__full_snp = full_snapshot

        self.config = configmanager
        self.__state = backupstate
        self.__state.clear_backup_properties()

        self.__profilename = self.config.getProfileName()
        self.__state.set_profilename(self.__profilename)

        self.__snpman = None

        self.__snapshot = None

        self.__fam_target_hdl = fam.get_fam_target_handler_facade_instance()

        self.logger.debug("Instance of BackupProfileHandler created.")

    def do_hook(self, hookname):
        """Runs scripts optionally defined in section 'hooks' of
        configuration file. Currently defined are
        * pre-backup - before preparation of backup starts
        * post-backup - after completion and finishing of backup
        
        """
        #LP #173490
        import commands
        hooks = None
        if self.config.has_option('hooks', hookname):
            hooks = str(self.config.get('hooks', hookname)).split(",")
    
        if hooks is not None:
            self.logger.info(_("Running of hooks: %s") % hookname)
            for hook in hooks:
                result = commands.getstatusoutput(hook)
                if( 0 != result[0]):
                    raise exceptions.HookedScriptError(\
                      "Hook %s returned error: '%s' (exit code=%s)" % (hookname,
                                                                    result[1],
                                                                    result[0]))

    def prepare(self):
        self.logger.info(_("Preparation of backup process"))
        _uri = self.config.get_destination_path()
        try:
            self.__fam_target_hdl.set_destination(_uri)
            self.__fam_target_hdl.set_configuration_ref(self.config)
            self.__fam_target_hdl.set_use_mainloop(use = True)
            self.__fam_target_hdl.initialize()
        except exceptions.FileAccessException:
            self.__fam_target_hdl.terminate()
            raise

        self.__check_target()

    def process(self):
        """Runs the whole backup process:
        
        1. check pre-conditions
        3. purge snapshots (if configured)
        4. open new snapshot containing common metadata (full or incr.
            depending on existing one, base, settings etc.)
        5. fill new snapshot (with packages list, include lists, exclude lists,
            size prediction)
        6. commit new snapshot to disk (creates the actual tar archive and
            writes everything into the snapshot directory).
        """
        assert self.__fam_target_hdl.is_initialized()

        self.__snpman = SnapshotManager(self.__fam_target_hdl.query_mount_uri())
        
        # get basic informations about new snapshot
        self.__state.set_state('prepare')
        (snppath, base) = self.__retrieve_basic_infos(force_full_snp = self.__full_snp)

        # Create a new snapshot
        self.__snapshot = snapshot.Snapshot(snppath)
        self.logger.info(_("Snapshot '%(name)s' is being made.")
                         % {'name' :str(self.__snapshot)})

        # Set the base file
        if base is not None:
            if self.__snapshot.isfull():
                self.logger.debug("Base is not being set for this full snapshot.")
            else:
                self.logger.info(_("Setting Base to '%(value)s'.") % {'value' : str(base)})
                self.__snapshot.setBase(base.getName())

        # Backup list of installed packages
        _packagecmd = "dpkg --get-selections"
        if self.config.has_option("general", "packagecmd"):
            _packagecmd = self.config.get("general", "packagecmd")
        if _packagecmd:
            try:
                self.logger.info(_("Setting packages File."))
                s = os.popen(_packagecmd)
                pkg = s.read()
                s.close()
                self.__snapshot.setPackages(pkg)
            except Exception, _exc:
                self.logger.warning(_("Problem when setting the packages list: ") + str(_exc))

        # set Excludes
# TODO: improve handling of Regex containing ',' (delimiter); currently this will crash
        self.logger.info(_("Setting Excludes File."))
        if self.config.has_option("exclude", "regex"):
            gexclude = str(self.config.get("exclude", "regex")).split(",")
        else :
            gexclude = ""
        self.__snapshot.setExcludes(gexclude)

        _compr = self.config.get_compress_format()
        self.logger.info(_("Setting compression format to `%s`") % _compr)
        self.__snapshot.setFormat(_compr)

        if self.config.has_option("general", "splitsize"):
            _chunks = int(self.config.get("general", "splitsize"))
            self.__snapshot.setSplitedSize(_chunks)
            if _chunks:
                self.logger.info(_("Setting size of archive chunks to %s")\
                            % util.get_humanreadable_size_str(size_in_bytes = (_chunks * 1024),
                                                              binary_prefixes = True))

        # set followlinks
        self.__snapshot.setFollowLinks(self.config.get_followlinks())
        if self.__snapshot.isFollowLinks():
            self.logger.info(_("Option 'Follow symbolic links' is enabled."))
        else:
            self.logger.info(_("Option 'Follow symbolic links' is disabled."))

        self.__collect_files()

        _publish_progress = True
        if self.__dbus_conn is None:
            _publish_progress = False
        _supports_publish = self.__fam_target_hdl.get_supports_publish()

        self.logger.info(_("Snapshot is being committed"))
        self.__state.set_state('start')
#        self.__state.set_state('commit')
        self.__snapshot.commit(self.__fam_target_hdl, _publish_progress, _supports_publish)

#TODO: add state purging
        # purge
        purge = None
        if self.config.has_option("general", "purge"):
            purge = self.config.get("general", "purge")
        if purge is not None:
            try:
                self.__snpman.purge(purge, self.__snapshot.getName()) # do not purge created snapshot
            except exceptions.SBException, sberror:
                self.logger.error(_("Error while purging old snapshots: %s") % sberror)

        self.logger.info(_("Backup process finished."))
        self.__state.set_state('finish')

    def __collect_files(self):
        """Fill snapshot's include and exclude lists and retrieve some information
        about the snapshot (uncompressed size, file count).
        """
        self.logger.info(_("Inspect file system and collect file infos"))
        _collector = self.__create_collector_obj()
        _collector.collect_files()
        _stats = _collector.get_stats()
        _snpsize = _stats.get_size_payload() + _stats.get_size_overhead(size_per_item = constants.TAR_BLOCKSIZE)

        self.__state.set_space_required(_snpsize)
        self.__snapshot.set_space_required(_snpsize)
        _sizefs, _freespace = self.__fam_target_hdl.query_dest_fs_info()

        _snpsize_hr = util.get_humanreadable_size_str(size_in_bytes = _snpsize, binary_prefixes = True)
        self.logger.info(_("Summary of backup"))
        self.logger.info(_("Number of directories: %s.") % _stats.get_count_dirs())
        self.logger.info(_("Total number of files: %s.") % _stats.get_count_files_total())
        self.logger.info(_("Number of symlinks: %s.") % _stats.get_count_symlinks())
        self.logger.info(_("Number of files included in snapshot: %s.") % _stats.get_count_files_incl())
        self.logger.info(_("Number of new files (also included): %s.") % _stats.get_count_files_new())
        self.logger.info(_("Number of files skipped in incremental snapshot: %s.") % _stats.get_count_files_skip())
        self.logger.info(_("Number of items forced to be excluded: %s.") % _stats.get_count_items_excl_forced())
        self.logger.info(_("Number of items to be excluded by config: %s.") % _stats.get_count_items_excl_config())
        self.logger.info(_("Maximum free size required is '%s'.") % _snpsize_hr)

        if _freespace == constants.FREE_SPACE_UNKNOWN:
            self.logger.warning("Unable to query available space on target: Operation not supported")
        else:
            _freespace_hr = util.get_humanreadable_size_str(size_in_bytes = _freespace, binary_prefixes = True)
            self.logger.info(_("Available disk size is '%s'.") % _freespace_hr)
            if _freespace <= _snpsize:
                raise exceptions.SBException(_("Not enough free space in the target directory for the planned backup (free: %(freespace)s, required: %(neededspace)s).")\
                                               % { 'freespace' : _freespace_hr, 'neededspace' : _snpsize_hr})

    def __create_collector_obj(self):
        """Factory method that returns instance of `FileCollector`.
        """
        _eff_local_dest_path = self.__fam_target_hdl.get_eff_path()
        _configfac = filecollect.FileCollectorConfigFacade(self.config, _eff_local_dest_path)
        _collect = filecollect.FileCollector(self.__snapshot, _configfac)
        if not self.__snapshot.isfull():
            _base = self.__snapshot.getBaseSnapshot()
            _basesnar = _base.getSnapshotFileInfos().get_snapfile_obj()
            _collect.set_parent_snapshot(_basesnar)
        return _collect

    def __copylogfile(self):
# TODO: we should flush the log file before copy!
        _op = fam.get_file_operations_facade_instance()

        if not self.__fam_target_hdl.is_initialized():
            self.logger.warning(_("Unable to copy log. File access is not initialized."))
        else:
            if self.__snapshot is not None:
                logf_src = self.config.get_current_logfile()
                if logf_src is None:
                    self.logger.warning(_("No log file specified."))
                else:
                    logf_name = _op.get_basename(logf_src)
                    logf_target = _op.joinpath(self.__snapshot.getPath(),
                                                logf_name)

                    if _op.path_exists(logf_src):
                        try:
                            _op.copyfile(logf_src, logf_target)
                        except exceptions.CopyFileAttributesError:
                            self.logger.warning(_("Unable to change permissions for file '%s'.")\
                                            % logf_target)
                        except (OSError, IOError), error:
                            self.logger.warning(_("Unable to copy log file: %s") % error)
                    else :
                        self.logger.warning(_("Unable to find logfile to copy into snapshot."))
            else:
                self.logger.warning(_("No snapshot to copy logfile."))

    def cancel(self):
        if self.__snapshot is not None:
            self.__snpman.remove_snapshot_forced(self.__snapshot)

        self.__fam_target_hdl.terminate()
        self.logger.info(_("Processing of profile was canceled on user request\n"))
        _excode = constants.EXCODE_SUCCESS
        return _excode

    def finish(self, error = None):
        """End SBackup session :
        
        - copy the log file into the snapshot dir
        
        Might be called multiple times.
        
        :note: When this method is called no exceptions were raised during the backup.
        """
        self.__copylogfile()
        self.__fam_target_hdl.terminate()

        if error is None:
            self.logger.info(_("Processing of profile successfully finished (no errors)\n"))
            _excode = constants.EXCODE_SUCCESS
        else:
            err_str = str(error)
            if err_str == "":
                err_str = str(type(error))
            self.logger.info(_("Processing of profile failed with error: %s\n") % err_str)
            _excode = constants.EXCODE_BACKUP_ERROR

        return _excode

    def __check_target(self):
        assert self.__fam_target_hdl.is_initialized(), "File access manager not initialized"

# TODO: Improve handling of original and modified target paths. Support display names for state (improved user interaction)
        _target_display_name = self.__fam_target_hdl.query_dest_display_name()
        self.__state.set_target(_target_display_name)
        self.logger.info(_("Backup destination: %s") % _target_display_name)

        # Check if the target dir exists, but Do not create any directories. 
        if not self.__fam_target_hdl.dest_path_exists():
            self.logger.warning(_("Unable to find destination directory."))
            self.__state.set_state('target-not-found')

            if self.__use_indicator and self.__dbus_conn is not None:
                _time = 0
                _retry = constants.RETRY_UNKNOWN

                while (_time <= constants.TIMEOUT_RETRY_TARGET_CHECK_SECONDS):
                    time.sleep(constants.INTERVAL_RETRY_TARGET_CHECK_SECONDS)
                    _time = _time + constants.INTERVAL_RETRY_TARGET_CHECK_SECONDS
#TODO: put the get_retry_target.. into State?
                    _retry = self.__dbus_conn.get_retry_target_check()
                    if _retry == constants.RETRY_FALSE:
                        raise exceptions.BackupCanceledError

                    elif _retry == constants.RETRY_TRUE:
                        if self.__fam_target_hdl.dest_path_exists():
                            pass
                        else:
                            self.logger.warning(_("Unable to find destination directory even after retry."))
                            raise exceptions.SBException(_("Target directory '%(target)s' does not exist.")\
                                            % {"target" : _target_display_name})
                        break
                    else:
                        pass

            else:
                raise exceptions.SBException(_("Target directory '%(target)s' does not exist.")\
                                % {"target" : _target_display_name})

        try:
            self.__fam_target_hdl.test_destination()
        except exceptions.FileAccessException, error:
            self.logger.error(_("Unable to access destination: %s") % (error))
            raise error

    def __retrieve_basic_infos(self, force_full_snp = False):
        """Retrieves basic informations about the snapshot that is going
        to be created. This informations include:
        1. the path of the new snapshot
        2. the base of the new snapshot
        
        :param listing: a list of snapshots
        
        :return: the determined `snppath` and `base`
        :rtype: a tuple
        
        """
        _fop = fam.get_file_operations_facade_instance()
        # Get the list of snapshots that matches the latest snapshot format
        listing = self.__snpman.get_snapshots()
        agelimit = int(self.config.get("general", "maxincrement"))
        increment = False

        base = None
        if (len(listing) == 0) or (force_full_snp is True):
            increment = False
        else:
            # we got some snaphots 
            # we search for the last full 
            base = listing[0]
            if listing[0].isfull() :  # Last backup was full backup
                self.logger.debug("Last (%s) was a full backup" % listing[0].getName())
                d = listing[0].getDate()
                age = (datetime.date.today() - datetime.date(d["year"], d["month"], d["day"])).days
                if  age < agelimit :
                    # Less than maxincrement days passed since that -> make an increment
                    self.logger.info("Last full backup is %i days old < %s -> make inc backup" % (age, agelimit))
                    increment = True
                else:
                    self.logger.info("Last full backup is %i days old > %s -> make full backup" % (age, agelimit))
                    increment = False      # Too old -> make full backup
            else: # Last backup was an increment - lets search for the last full one
                self.logger.debug(" Last snapshot (%s) was incremental. Lookup of latest full snapshot." % listing[0].getName())
                for i in listing :
                    if i.isfull():
                        d = i.getDate()
                        age = (datetime.date.today() - datetime.date(d["year"], d["month"], d["day"])).days
                        if  age < agelimit :
                            # Last full backup is fresh -> make an increment
                            self.logger.info("Last full backup is fresh (%d days old )-> make an increment" % age)
                            increment = True
                        else: # Last full backup is old -> make a full backup
                            self.logger.info("Last full backup is old -> make a full backup")
                            increment = False
                        break
                else:
                    self.logger.info("No full backup found -> lets make a full backup to be safe")
                    increment = False

        # Determine and create backup target directory
        hostname = socket.gethostname()
        snpname = "%s.%s" % (datetime.datetime.now().isoformat("_").replace(":", "."),
                             hostname)
        if increment is True:
            snpname = "%s.inc" % snpname
        else:
            snpname = "%s.ful" % snpname

        tdir = self.__fam_target_hdl.get_snapshot_path(snpname)
        self.logger.debug("Snapshot path: %s" % tdir)

        return (tdir, base)



#   Simple Backup - Base class for Gnome applications
#
#   Copyright (c)2009-2010: Jean-Peer Lorenz <peer.loz@gmx.net>
#   Copyright (c)2007: Ouattara Oumar Aziz <wattazoum@gmail.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#   The code is based on code originally written by Dave Reed (12/15/2002)


from gettext import gettext as _
import os.path

import gtk.glade
import gobject
import glib


PROGRESSBAR_PULSE_INTERVAL = 150


def search_file(filename, search_path):
    """Given a search path, find file
    """
    file_found = False
    paths = search_path.split(os.pathsep)
    _result = None
    for _path in paths:
        _search = os.path.join(_path, filename)
        if os.path.exists(_search):
            file_found = True
            break
    if file_found:
        _result = os.path.abspath(_search)
    return _result


class GladeWindow(object):
    '''A base class for displaying a GUI developed with Glade; create
    a subclass and add any callbacks and other code; the derived class
    __init__ method needs to call GladeWindow.__init__; callbacks that
    start with on_ are automatically connected'''

    def __init__(self, gladefile, widget_list, handlers, root, parent = None,
                  pull_down_dict = None):

        '''
        __init__(self, filename, top_window, widget_list, pull_down_dict=None):

        filename: filename of the .glade file
        top_window: the glade name of the top level widget (this will then
           be accessible as self.top_window)
        widget_list: a list of glade names; the dictionary self.widgets
           will be created that maps these name to the actual widget object
        pull_down_dict: a dictionary that maps combo names to a tuple of
            strings to put in the combo
            
        @todo: JPL: I don't want to change too much in this release but for\
               later releases: Create the XML external and give it to the\
               widget classes to avoid multiple instantiations!
        '''
        try:
            search_path = GladeWindow.search_path
        except:
            search_path = './'

        fname = search_file(gladefile, search_path)
        self.xml = gtk.glade.XML(fname, root = root, domain = 'sbackup')

        # connect callbacks
        self.cb_dict = {}
        for f in handlers:
            self.cb_dict[f] = getattr(self, f)
        self.xml.signal_autoconnect(self.cb_dict)

        self.widgets = {}
        for widget in widget_list:
            self.widgets[widget] = self.xml.get_widget(widget)

        if pull_down_dict is not None:
            for widget, lst in pull_down_dict.items():
                self.widgets[widget].set_popdown_strings(lst)

        # set attribute for top_window so it can be accessed as self.top_window
        self.top_window = None

        # window to show when this one is hidden
# TODO: What is this for? Should be handled outside this class!
        self.prev_window = None

        # initialize callback func
        self.cb_func = None

        self.parent = parent


    def set_search_path(cls, path):

        '''set the search path for looking for the .glade files'''

        cls.search_path = path

    set_search_path = classmethod(set_search_path)

    def set_top_window(self, top_window):

        '''set_top_window(self, top_window):

        notebook pages that are in containers need to be able to change
        their top window, especially so the dialog is set_transient_for
        the actual main window
        '''
        self.top_window = top_window

#    def set_callback_function(self, cb_func, *cb_args, **cb_kwargs):
#
#        '''set_callback_function(cb_func, *cb_args, **cb_kwargs):
#
#        stores the cb_func and its cb_args and cb_kwargs
#        '''
#        self.cb_func = cb_func
#        self.cb_args = cb_args
#        self.cb_kwargs = cb_kwargs

    def show(self, center = 1, prev_window = None, *args):

        '''show(self, center=1, prev_window=None, *args):

        display the top_window widget
        '''

        if prev_window is not None:
            self.prev_window = prev_window
#        if center:
#            self.top_window.set_position(gtk.WIN_POS_CENTER_ALWAYS)
#        else:
#            self.top_window.set_position(gtk.WIN_POS_NONE)

        self.top_window.show()

    def hide(self):
        '''hide(self):

        hides the current window, shows self.prev_window
        if self.cb_func is not None, it is called with its cb_args
        and cb_kwargs
        '''

        self.top_window.hide()
        if self.prev_window is not None:
            self.prev_window.show()
#        if self.cb_func is not None:
#            self.cb_func(*self.cb_args, **self.cb_kwargs)
        if self.prev_window is None:
            gtk.main_quit()

    def _show_infomessage(self, message_str, boxtitle = "",
                               headline_str = "", secmsg_str = ""):
        self._show_message(gtk.MESSAGE_INFO, message_str, boxtitle, headline_str, secmsg_str)

    def _show_errmessage(self, message_str, boxtitle = "",
                               headline_str = "", secmsg_str = ""):
        self._show_message(gtk.MESSAGE_ERROR, message_str, boxtitle, headline_str, secmsg_str)

    def _show_message(self, msgtype, message_str, boxtitle = "",
                            headline_str = "", secmsg_str = ""):
        """Shows message box using markup.
        
        @attention: gtk.MessageDialog is not fully thread-safe so it is
                    put in a critical section, here!
        """
        gtk.gdk.threads_enter()
        dialog = gtk.MessageDialog(
                    flags = gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                    type = msgtype,
                    buttons = gtk.BUTTONS_CLOSE)
        if boxtitle.strip() != "":
            dialog.set_title(boxtitle)

        _hdl = headline_str.strip(" \n\t")
        if _hdl != "":
            _hdl = "<b>%s</b>\n\n" % (glib.markup_escape_text(_hdl))
        _msg = "%s%s" % (_hdl, glib.markup_escape_text(message_str))
        dialog.set_markup(_msg)

        # an optional secondary message is added
        _sec = secmsg_str.strip(" \n\t")
        if _sec != "":
            _sec = "<small>%s</small>" % (glib.markup_escape_text(_sec))
            dialog.format_secondary_markup(_sec)

        # show the message box and destroy it afterwards
        dialog.run()
        dialog.destroy()
        gtk.gdk.threads_leave()

    def _show_destination_error(self, error):
        _sec_msg = _("Please make sure the missing directory exists (e.g. by mounting an external disk) or change the specified target.")
        self._show_errmessage(message_str = _("While accessing the backup destination following error occurred:\n\n%s.") % error,
                boxtitle = _("Simple Backup error"),
                headline_str = _("Unable to access backup destination"),
                secmsg_str = _sec_msg)



class ProgressbarMixin(object):
    """Extends a class by functionality for showing some action using
    a pulsing progressbar.
    """

    def __init__(self, progressbar, hide_when_stopped = True):
        """Default constructor. The progressbar that should be used must be
        given as parameter. If the optional parameter 'hide_when_stopped' is
        set, the progressbar is only showed if it is active and will be
        hidden afterwards.
        
        @param progressbar: a progressbar object
        @type progressbar: GtkProgressBar
        @param hide_when_stopped: if true, the progressbar is hidden when
                                  not active
        @type hide_when_stopped: Boolean
        
        @return: None
        """
        self._progressbar = progressbar
        self.__pulsetimer_id = None
        self.__hide_when_stopped = hide_when_stopped

    def _init_pulse(self):
        """Convenience method for initializing the progressbar during
        the application's startup.
        """
        self.__pulsetimer_id = None
        if self.__hide_when_stopped:
            self._progressbar.hide()
        self._progressbar.set_text("")
        self._progressbar.set_fraction(0.0)

    def _start_pulse(self):
        """Calling this method shows up the progressbar and starts
        the pulsing.
        
        @return: None
        """
        if self.__pulsetimer_id is not None:
            gobject.source_remove(self.__pulsetimer_id)
            self.__pulsetimer_id = None
        self._progressbar.set_text("")
        self._progressbar.set_fraction(0.0)
        if self.__hide_when_stopped:
            self._progressbar.show()
        self.__pulsetimer_id = gobject.timeout_add(PROGRESSBAR_PULSE_INTERVAL, self.__pulse)

    def _stop_pulse(self):
        """Calling this method stops the progressbar.
        
        @return: False
        @rtype:  Boolean
        """
        if self.__pulsetimer_id is not None:
            gobject.source_remove(self.__pulsetimer_id)
        self.__pulsetimer_id = None
        if self.__hide_when_stopped:
            self._progressbar.hide()
        self._progressbar.set_text("")
        self._progressbar.set_fraction(0.0)
        return False

    def __pulse(self):
        """Private helper method that actually performs the pulsing.
        
        @return: True
        @rtype:  Boolean        
        """
        self._progressbar.pulse()
        return True



#   Simple Backup - Target file access management using GIO/GVFS
#
#   Copyright (c)2010: Jean-Peer Lorenz <peer.loz@gmx.net>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#


from gettext import gettext as _
import types

from sbackup.fs_backend import _gio_utils as gio_utils
from sbackup.fs_backend._gio_utils import GioOperations

from sbackup.util import exceptions
from sbackup.util import interfaces
from sbackup.util import pathparse
from sbackup.util import log


class GioTargetHandler(interfaces.ITargetHandler):

    def __init__(self):
        """
        Constructor
        :param configuration:
         
        only mounted path is Target (currently)        
        """
        interfaces.ITargetHandler.__init__(self)

        self._logger = log.LogFactory.getLogger()

        # destination/target specific instances
        self._dest = None
        self._dest_mount_hdl = gio_utils.GioMountHandler()
        self._use_mainloop = False
        self._configuration = None
        self._in_progress = False  # must be set to False when start mount process (lock)
        self._initialize_callback = None
        self._terminate_callback = None
        self._is_initialized = False

    def set_use_mainloop(self, use):
        if not isinstance(use, types.BooleanType):
            raise TypeError("Expected boolean type")
        self._use_mainloop = use

    def set_initialize_callback(self, func):
        self._logger.debug("set_initialize_callback: %s" % (str(func)))
        self._initialize_callback = func

    def set_terminate_callback(self, func):
        self._logger.debug("set_terminate_callback: %s" % (str(func)))
        self._terminate_callback = func

    def set_destination(self, path):
        _dest_obj = pathparse.UriParser()
        _dest_obj.set_and_parse_uri(uri = path)
        self._dest = _dest_obj
        self._dest_mount_hdl.set_uri(self._dest) # actually a UriParser

    def get_destination(self):
        _path = self._dest.uri
        return _path

    def is_local(self):
        _loc = self._dest.is_local()
        return _loc

    def get_supports_publish(self):
        """Current implementations of gvfs-fuse causes issues when using
        tar's checkpoint action in conjunction with multiple volumes.
        """
        _res = False
        if self.is_local():
            _res = True
        return _res

    def get_use_iopipe(self):
        _res = True
        return _res

    def set_configuration_ref(self, configuration):
        self._configuration = configuration

    def initialize(self):
        self._logger.info(_("Initializing GIO File Access Manager."))
        if self._dest is None:
            raise TypeError("No destination given")

#TODO: skip initialization or raise exception?        
        if self._is_initialized is True:
            raise AssertionError("GIO File Access Manager is already initialized.")

        self._mount_destination()

    def is_initialized(self):
        return self._is_initialized

    def terminate(self):
        self._logger.info(_("Terminating GIO File Access Manager."))
        if self._in_progress is not False:
            raise AssertionError("Another mount process is still in progress")
        self._in_progress = True

        if self._is_initialized is True:
            # set `use_mainloop` again for the case it was changed in the meantime
            self._dest_mount_hdl.use_own_mainloop(self._use_mainloop)
            try:
                self._dest_mount_hdl.umount()
            except exceptions.RemoteUmountFailedError, error:
                self._logger.error("Unable to umount `%s`: %s" % (self._dest.query_display_name(), error))

        else:
            self._logger.warning("GIO File Access Manager is not initialized. Nothing to do.")
            self._logger.debug("Calling callbacks anyway")
            self._umount_cb(None)

    def _mount_destination(self):
        """Mounts the destination specified in configuration and modifies the target value
        in configuration accordingly.
        """
        if self._in_progress is not False:
            raise AssertionError("Another mount process is still in progress")

        self._in_progress = True  # set lock

        self._dest_mount_hdl.use_own_mainloop(self._use_mainloop)
        self._dest_mount_hdl.set_callbacks(mount = self._mount_cb,
                                                      umount = self._umount_cb)
        self._dest_mount_hdl.mount()

    def _mount_cb(self, error):
        """Callback method that gets called when mounting is finished.
        Takes errors occurred during the mount process as parameter.
        """
        self._in_progress = False # release lock

        if self._initialize_callback is not None:
            self._logger.debug("Calling additional callback in gio_fam._mount_cb: %s" % self._initialize_callback)
            self._initialize_callback(error)
        else:
            if error is not None:
                raise exceptions.FileAccessException("Unable to mount: %s" % error)

        if error is None:
            self._is_initialized = True

    def _umount_cb(self, error):
        self._in_progress = False # release lock
        if self._terminate_callback is not None:
            self._logger.debug("Calling additional callback in gio_fam: %s" % self._terminate_callback)
            self._terminate_callback(error)

        self._is_initialized = False

    def get_eff_path(self):
        _path = self.query_mount_uri()
        _eff_path = GioOperations.get_eff_path(_path)
        return _eff_path

    def get_eff_fullpath(self, *args):
        _base = GioOperations.normpath(*args)
        _eff_path = self.get_eff_path()
        self._logger.debug("Effective path: `%s`" % _eff_path)
        self._logger.debug("Base path: `%s`" % _base)
        _res = None
        if _eff_path is not None:
            _res = GioOperations.normpath(_eff_path, _base)
        return _res

    def query_dest_fs_info(self):
        (_size, _free) = self._dest_mount_hdl.query_fs_info()
        return (_size, _free)

    def query_dest_display_name(self):
        return self._dest.query_display_name()

    def query_mount_uri(self):
        return self._dest.query_mount_uri()

    def dest_path_exists(self):
        """The effective path denotes the local mountpoint of the actual remote or local target.
        It is required in order to give it to TAR as parameter (tar does not support gio).
        It is checked using GIO and native access functions.
        """
        _path = self.query_mount_uri()
        _res_gio = GioOperations.path_exists(_path)
        return _res_gio

    def test_destination(self):
        self._dest_mount_hdl.test_path()

    def get_snapshot_path(self, snpname):
        _base = self.query_mount_uri()
        _snppath = GioOperations.normpath(_base, snpname)
        return _snppath



#   Simple Backup - Base class for Gnome applications (deprecated)
#
#   Copyright (c)2010: Jean-Peer Lorenz <peer.loz@gmx.net>
#   Copyright (c)2007: Ouattara Oumar Aziz <wattazoum@gmail.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#


import os
import gtk.glade
import gnome
import gnome.ui # required by GnomeAppBar


def search_file(filename, search_path):

    """Given a search path, find file
    """

    file_found = 0
    paths = search_path.split(os.pathsep)

    for path in paths:
        if os.path.exists(os.path.join(path, filename)):
            file_found = 1
            break

    if file_found:
        return os.path.abspath(os.path.join(path, filename))
    else:
        return None

#----------------------------------------------------------------------

class GladeGnomeApp(object):

    '''A base class for displaying a GUI developed with Glade; create
    a subclass and add any callbacks and other code; the derived class
    __init__ method needs to call GladeWindow.__init__; callbacks that
    start with on_ are automatically connected'''

    #----------------------------------------------------------------------

    @classmethod
    def set_search_path(cls, path):
        '''set the search path for looking for the .glade files'''

        cls.search_path = path

    #----------------------------------------------------------------------

    def __init__(self, app_name, app_version, filename, top_window, widget_list, handlers,
                 pull_down_dict = None):

        '''
        __init__(self, app_name, app_version, filename, top_window, widget_list, pull_down_dict=None):
        
        app_name, app_version : application name and version 
        filename: filename of the .glade file
        top_window: the glade name of the top level widget (this will then
           be accessible as self.top_window)
        widget_list: a list of glade names; the dictionary self.widgets
           will be created that maps these name to the actual widget object
        pull_down_dict: a dictionary that maps combo names to a tuple of
            strings to put in the combo
        '''
        self.widget_list = widget_list

        try:
            search_path = GladeGnomeApp.search_path
        except:
            search_path = './'

        gnome.program_init(app_name, app_version)

        fname = search_file(filename, search_path)
        self.xml = gtk.glade.XML(fname, domain = 'sbackup')

        # prepare callbacks
        self.cb_dict = {}
        for f in handlers:
            self.cb_dict[f] = getattr(self, f)

        self.widgets = {}
        for w in self.widget_list:
            self.widgets[w] = self.xml.get_widget(w)

        if pull_down_dict is not None:
            for w, l in pull_down_dict.items():
                self.widgets[w].set_popdown_strings(l)

        # set attribute for top_window so it can be accessed as self.top_window
        self.top_window = self.xml.get_widget(top_window)

        # window to show when this one is hidden
        self.prev_window = None

        # initialize callback func
        self.cb_func = None

    def set_top_window(self, top_window):

        '''set_top_window(self, top_window):
        
        notebook pages that are in containers need to be able to change
        their top window, especially so the dialog is set_transient_for
        the actual main window
        '''

        self.top_window = top_window

#    def set_callback_function(self, cb_func, *cb_args, **cb_kwargs):
#
#        '''set_callback_function(cb_func, *cb_args, **cb_kwargs):
#        
#        stores the cb_func and its cb_args and cb_kwargs
#        '''
#        self.cb_func = cb_func
#        self.cb_args = cb_args
#        self.cb_kwargs = cb_kwargs

    def show(self, center = 1, prev_window = None, *args):

        '''show(self, center=1, prev_window=None, *args):
        
        display the top_window widget
        '''

        if prev_window is not None:
            self.prev_window = prev_window
#        if center:
#            self.top_window.set_position(gtk.WIN_POS_CENTER_ALWAYS)
#        else:
#            self.top_window.set_position(gtk.WIN_POS_NONE)
        self.top_window.show()

    def hide(self):

        '''hide(self):
        
        hides the current window, shows self.prev_window
        if self.cb_func is not None, it is called with its cb_args
        and cb_kwargs
        '''

        self.top_window.hide()
        if self.prev_window is not None:
            self.prev_window.show()
#        if self.cb_func is not None:
#            self.cb_func(*self.cb_args, **self.cb_kwargs)
        if self.prev_window is None:
            gtk.main_quit()



#   Simple Backup - DBus System Service
#
#   Copyright (c)2009-2010: Jean-Peer Lorenz <peer.loz@gmx.net>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# This code is based on examples published by
#   Red Hat Inc. <http://www.redhat.com/>
#   Collabora Ltd. <http://www.collabora.co.uk/>
#
"""
:mod:`sbackup.dbus_service` -- DBus System Service
====================================================

.. module:: dbus_service
   :synopsis: Provides DBus System Service.
.. moduleauthor:: Jean-Peer Lorenz <peer.loz@gmx.net>

"""

import os
import uuid
import optparse

import gobject

import dbus.service
import dbus.mainloop.glib

from sbackup.pkginfo import Infos
from sbackup.util import log
from sbackup.util import constants
from sbackup.util import exceptions
from sbackup.util.system import drop_privileges
from sbackup.util.system import pid_exists


#TODO: Further split of object's responsibility: sender object/receiver 
#TODO: Add module dbus_constants?
__START = "start"
__STOP = "stop"
__RESTART = "restart"
__valid_args = [__START, __STOP, __RESTART]


class SBDbusException(dbus.DBusException):
    _dbus_error_name = constants.DBUS_EXCEPTION


class SBackupdDBusObject(dbus.service.Object):

    def __init__(self, system_bus, object_path):
        dbus.service.Object.__init__(self, system_bus, object_path)
        self._system_bus = system_bus

        self.__backup_pid = constants.PID_UNKNOWN
        self.__retry_target_check = constants.RETRY_UNKNOWN
        self.__target = constants.TARGET_UNKNOWN
        self.__profilename = constants.PROFILE_UNKNOWN
        self.__space_required = constants.SPACE_REQUIRED_UNKNOWN

    @dbus.service.signal(constants.DBUS_INTERFACE)
    def progress_signal(self, checkpoint):
        pass

    @dbus.service.signal(constants.DBUS_INTERFACE)
    def event_signal(self, event, urgency):
        pass

    @dbus.service.signal(constants.DBUS_INTERFACE)
    def error_signal(self, error):
        pass

    @dbus.service.signal(constants.DBUS_INTERFACE)
    def targetnotfound_signal(self):
        pass

    @dbus.service.signal(constants.DBUS_INTERFACE)
    def alreadyrunning_signal(self):
        pass

    @dbus.service.method(constants.DBUS_INTERFACE,
                         in_signature = 'ss', out_signature = 'b')
    def emit_event_signal(self, event, urgency):
        self.event_signal(event, urgency)
        return True

    @dbus.service.method(constants.DBUS_INTERFACE,
                         in_signature = 's', out_signature = 'b')
    def emit_error_signal(self, error):
        self.error_signal(error)
        return True

    @dbus.service.method(constants.DBUS_INTERFACE,
                         in_signature = '', out_signature = 'b')
    def emit_targetnotfound_signal(self):
        self.targetnotfound_signal()
        return True

    @dbus.service.method(constants.DBUS_INTERFACE,
                         in_signature = '', out_signature = 'b')
    def emit_alreadyrunning_signal(self):
        self.alreadyrunning_signal()
        return True

    @dbus.service.method(constants.DBUS_INTERFACE,
                         in_signature = 'n', out_signature = 'b')
    def set_retry_target_check(self, retry):
        """
        Valid values are:  -1 unknown
                            0 False (do not retry)
                            1 True (retry)
        """
        self.__retry_target_check = retry
        return True

    @dbus.service.method(constants.DBUS_INTERFACE,
                         in_signature = '', out_signature = 'n')
    def get_retry_target_check(self):
        return self.__retry_target_check

    @dbus.service.method(constants.DBUS_INTERFACE,
                         in_signature = 'i', out_signature = 'b')
    def set_backup_pid(self, pid):
        """
        """
        self.__backup_pid = pid
        return True

    @dbus.service.method(constants.DBUS_INTERFACE,
                         in_signature = '', out_signature = 'i')
    def get_backup_pid(self):
        """
        """
        return self.__backup_pid

    @dbus.service.method(constants.DBUS_INTERFACE,
                         in_signature = 's', out_signature = 'b')
    def set_target(self, target):
        """
        """
        self.__target = target
        return True

    @dbus.service.method(constants.DBUS_INTERFACE,
                         in_signature = 's', out_signature = 'b')
    def set_profilename(self, profilename):
        """
        """
        self.__profilename = profilename
        return True

    @dbus.service.method(constants.DBUS_INTERFACE,
                         in_signature = 'x', out_signature = 'b')
    def set_space_required(self, space):
        """
        """
        self.__space_required = space
        return True

    @dbus.service.method(constants.DBUS_INTERFACE,
                         in_signature = '', out_signature = 's')
    def get_target(self):
        return self.__target

    @dbus.service.method(constants.DBUS_INTERFACE,
                         in_signature = '', out_signature = 's')
    def get_profilename(self):
        return self.__profilename

    @dbus.service.method(constants.DBUS_INTERFACE,
                         in_signature = '', out_signature = 'x')
    def get_space_required(self):
        return self.__space_required

    @dbus.service.method(constants.DBUS_INTERFACE,
                         in_signature = 's', out_signature = 'b')
    def emit_progress_signal(self, checkpoint):
        self.progress_signal(checkpoint)
        return True


class SBackupDbusConnectionObject(dbus.service.Object):

    def __init__(self, system_bus, object_path, mainloop, keep_alive = False):
        dbus.service.Object.__init__(self, system_bus, object_path)
        self.__logger = log.LogFactory.getLogger()
        self._system_bus = system_bus
        self._mainloop = mainloop
        self._listeners = {}
        self._keep_alive = keep_alive
        self._id = None

        self._set_id()
        self._initialize_timer()

    def _set_id(self):
        pid = str(os.getpid())
        oid = str(id(self))
        uid = str(uuid.uuid4())
        self._id = "%s:%s:%s" % (pid, oid, uid)

    def _initialize_timer(self):
        gobject.timeout_add(constants.INTERVAL_UNREGISTER_DEAD_CONN,
                            self._unregister_dead_connections)

    def _unregister_dead_connections(self):
        self.__logger.debug("Check for dead connections")
        for _client in self._listeners.keys():
            pid = self._listeners[_client][1]
            if not pid_exists(pid):
                self.__logger.info("connected client process does not exist. Removed.")
                self.unregister_connection(connection_id = _client)
        return True

    @dbus.service.method(constants.DBUS_INTERFACE,
                         in_signature = 'ss', out_signature = 's')
    def register_connection(self, name, pid):
        """
        """
        _res = ""
        self.__logger.debug("Register connection: %s (pid=%s)" % (name, pid))
        _id = str(uuid.uuid4())
        if self._listeners.has_key(_id):
            raise SBDbusException("The generated id is already in use. This is a bug :( Please report the issue to the developers.")
        else:
            self._listeners[_id] = (name, pid)
            _res = _id
        self.__logger.debug("INFO: My connections: %s" % str(self._listeners))
        return _res

    @dbus.service.method(constants.DBUS_INTERFACE,
                         in_signature = 's', out_signature = 'b')
    def unregister_connection(self, connection_id):
        """
        @todo: Add some kind of regular insanity check, e.g. by checking PIDs given
               when registering whether the listeners are still running.
        """
        _res = False
        self.__logger.debug("Unregister listener with id: %s" % connection_id)

        if not self._listeners.has_key(connection_id):
            raise SBDbusException("The given id is not registered.")
        else:
            del self._listeners[connection_id]
            _res = True
        if len(self._listeners) == 0:
            self.__logger.debug("Last listener was unregistered.")
            if self._keep_alive:
                self.__logger.info("But SBackupdDBusObject is kept alive.")
            else:
                self.__logger.info("SBackupdDBusObject is being terminated.")
                gobject.timeout_add(constants.DBUS_SERVICE_QUIT_PAUSE, self.quit)
        return _res

    @dbus.service.method(constants.DBUS_INTERFACE,
                         in_signature = '', out_signature = '')
    def quit(self):
        if self._mainloop:
            self._mainloop.quit()

    @dbus.service.signal(constants.DBUS_INTERFACE)
    def exit_signal(self):
        pass

    @dbus.service.method(constants.DBUS_INTERFACE,
                         in_signature = '', out_signature = 'b')
    def emit_exit_signal(self):
        self.exit_signal()
        return True

    @dbus.service.method(constants.DBUS_INTERFACE,
                         in_signature = '', out_signature = 's')
    def get_id(self):
        return self._id


class SBackupDBusService(object):
    """This is the DBUS service that provide basic signals.
    
    """
    def __init__(self, keep_alive = False):
        self._system_bus = None
        self._dbus_service = None
        self._remote_obj = None
        self._connection = None
        self._mainloop = None
        self._keep_alive = keep_alive

    def _initialize_mainloop(self):
        self._mainloop = gobject.MainLoop()

    def _initialize_dbus_service(self):
        if self._mainloop is None:
            raise AssertionError("ERR: Mainloop must be initialized before "\
                                 "starting the dbus service.")

        self._system_bus = dbus.SystemBus()
        self._dbus_service = dbus.service.BusName(\
                                      constants.DBUS_SERVICE,
                                      self._system_bus)
        self._remote_obj = SBackupdDBusObject(self._system_bus,
                                      constants.DBUS_OBJ_PATH)
        self._connection = SBackupDbusConnectionObject(self._system_bus,
                                      constants.DBUS_CONNECTION_OBJ_PATH,
                                      self._mainloop,
                                      keep_alive = self._keep_alive)

    def main(self):
        self._initialize_mainloop()
        self._initialize_dbus_service()
        self._mainloop.run()


def is_running():
    """
    :todo: Retrieve the PID of the running service!
    """
    res = False
    try:
        bus = dbus.SystemBus()
        dbus_object = bus.get_object('org.freedesktop.DBus',
                                     '/org/freedesktop/DBus')
        dbus_iface = dbus.Interface(dbus_object, 'org.freedesktop.DBus')
        services = dbus_iface.ListNames()
    except dbus.DBusException, error:
        raise exceptions.DBusException(str(error))
    else:
        services.sort()
        for service in services:
            if service == constants.DBUS_SERVICE:
                res = True
                break
    return res


def __launch_service(keep_alive = False):
    sbak_service = SBackupDBusService(keep_alive = keep_alive)
    sbak_service.main()


def __stop_service():
    try:
        _system_bus = dbus.SystemBus()
        remote_obj = _system_bus.get_object(constants.DBUS_SERVICE,
                                            constants.DBUS_CONNECTION_OBJ_PATH)
    except dbus.DBusException:
        remote_obj = None

    if remote_obj is not None:
        remote_obj.quit(dbus_interface = constants.DBUS_INTERFACE)


def parse_cmdline(argv):
    usage = "Usage: %prog " + __START + "|" + __STOP + "|" + __RESTART + " [options] (use -h or --help for more infos)"
    version = "%prog " + Infos.VERSION
    prog = constants.DBUSSERVICE_FILE

    parser = optparse.OptionParser(usage = usage, version = version, prog = prog)
    parser.add_option("--keep-alive",
              action = "store_true", dest = "keep_alive", default = False,
              help = "don't terminate DBus service after last client was unregistered")

    (options, args) = parser.parse_args(argv[1:])
    if len(args) != 1:
        parser.error("No command given")
    cmd = args[0]
    if cmd not in __valid_args:
        parser.error("Unknown command given")
    if cmd == "stop" and options.keep_alive:
        parser.error("Unable to stop the service and to keep it alive")

    return (cmd, options)


def run(args):
    try:
        _cmd, _options = parse_cmdline(argv = args)
        drop_privileges()
        os.nice(5)

        dbus.mainloop.glib.DBusGMainLoop(set_as_default = True)

        _running = is_running()
        if _cmd == "start":
            if _running is True:
                print "Simple Backup DBus service is already running"
            else:
                __launch_service(keep_alive = _options.keep_alive)
        elif _cmd == "stop":
            if _running is True:
                __stop_service()
            else:
                print "Simple Backup DBus service is not running"
        elif _cmd == "restart":
            if _running is True:
                __stop_service()
            __launch_service(keep_alive = _options.keep_alive)

        exitcode = constants.EXCODE_SUCCESS

    except dbus.DBusException, error:
        print "Error in Simple Backup DBus service:\n%s" % str(error)
        exitcode = constants.EXCODE_GENERAL_ERROR
    return exitcode



#   Simple Backup - local fuse plugin 
#
#   Copyright (c)2009-2010: Jean-Peer Lorenz <peer.loz@gmx.net>
#   Copyright (c)2007-2008: Ouattara Oumar Aziz <wattazoum@gmail.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from sbackup.fuse_plugins import pluginFAM
from sbackup.util import local_file_utils


class localFuseFAM (pluginFAM):
    """
    The localFuseFAM plugin do nothing when the mount method is called
    @author: Oumar Aziz Ouattara <wattazoum@gmail.com>
    @version: 1.0
    """
    def __init__(self):
        pluginFAM.__init__(self)

    def match_scheme(self, remoteSource):
        return remoteSource.startswith(local_file_utils.PATHSEP)

    def match_scheme_full(self, remoteSource):
        """
        Try to match the scheme of the remoteSource.
        @param remoteSource: The remote path
        @return: True if the scheme matches the one for this 
        @rtype: boolean
        """
        return remoteSource.startswith(local_file_utils.PATHSEP)

    def mount(self, source, mountbase):
        """
        Fake mounter to be used by the fuseFAM
        @param source: The remote path
        @param mountbase: The mount points base dir
        @return: The mount point complete path
        @rtype: str
        """
        self.logger.debug("Nothing to do for '%s'" % source)
        return (local_file_utils.PATHSEP, "", source.lstrip(local_file_utils.PATHSEP))

    def umount(self, mounteddir):
        """
        """
        self.logger.debug("No need to unmount %s" % mounteddir)



# This Python file uses the following encoding: utf-8
#
#   Simple Backup - package info
#
#   Copyright (c)2010,2013: Jean-Peer Lorenz <peer.loz@gmx.net>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#


import os
import socket
from gettext import gettext as _

from sbackup import util


class Infos(object):
    """Collects general informations about SBackup.
    
    `name` The application's full name
    `version` Full current version
    `description` some description
    
    @todo: Remove the smtpfrom info from here.
    
    """
    NAME = "Simple Backup Suite"
    VERSION = util.get_version_number()
    DESCRIPTION = _("This is a user friendly backup solution for common desktop needs. If you want to help by submitting bugs, by translating or anything else please visit our website.")
    WEBSITE = "https://launchpad.net/sbackup/"
    COPYRIGHT = "2005-2013 The Simple Backup authors"
    AUTHORS = [_("Maintainers:"),
               "Jean-Peer Lorenz <peer.loz@gmx.net>",
               "Oumar Aziz Ouattara <wattazoum@gmail.com>",
               "",
               _("Former maintainers:"),
               "Aigars Mahinovs",
               "Jonh Wendell",
               "",
               _("Contributors:"),
               "Martin Schaaf",
               "Florian Achleitner",
               "Andreas Sliwka",
               "Rogach (platon7pronko)",
               "Felix Griewald (tiiunder)",
               "Marc Deslauriers",
               "Thibault Godouet",
               "Anton Feenstra",
               "Derek Ditch",
               "Bernd Wurst",
               "Simon Déziel",
               "drakosha",
               "bytebybyte",
               "Marcel Stimberg",
               "Oliver Gerlich"
              ]
    TRANSLATORS = "\n".join([_("translator-credits"),
                             "",
                             "Japanese translation: Thanks to Hajime Mizuno"
                            ])


#TODO: remove snapshot version from here!
    SNPCURVERSION = "1.5"

#TODO: remove hostname and mailsuffix from here!
    hostname = socket.gethostname()
    if "." in hostname:
        mailsuffix = hostname
    else:
        mailsuffix = hostname + ".ext"
    SMTPFROM = _("SBackup Daemon <%(login)s@%(hostname)s>") % {
                    'login' : os.getenv("USERNAME"), 'hostname': mailsuffix }




#   Simple Backup - handling of preferences
#
#   Copyright (c)2010: Jean-Peer Lorenz <peer.loz@gmx.net>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
"""
:mod:`sbackup.util.prefs` -- handling of preferences
======================================================

.. module:: prefs
   :synopsis: Handling of preferences applied to all profiles (in contrast to profile settings)
.. moduleauthor:: Jean-Peer Lorenz <peer.loz@gmx.net>

"""


import glib
import gconf


from sbackup.util import structs


FS_BACKEND_GIO = "gio"
FS_BACKEND_FUSE = "fuse"

PREFS_FS_BACKEND = "/apps/sbackup/global-preferences/fs_backend"


_PREFS_DEFAULTS = { PREFS_FS_BACKEND : FS_BACKEND_GIO }
_PREFS_TYPES = { PREFS_FS_BACKEND : gconf.VALUE_STRING }


class Preferences(object):
    __metaclass__ = structs.Singleton

    def __init__(self):
        self._client = gconf.client_get_default ()

    def get(self, key):
        _type = _PREFS_TYPES[key]

        _value = self._get_value(key)
        if _value is None:
            # return default and set default in database (if possible)                        
            self._set_value(key, _PREFS_DEFAULTS[key])
            _value = _PREFS_DEFAULTS[key]

        return _value

    def _get_value(self, key):
        _type = _PREFS_TYPES[key]
        _value = None

        if _type == gconf.VALUE_STRING:
            try:
                _value = self._client.get_string(key)
                _value = _value.lower()
            except glib.GError, error:
                print "Error while getting gconf setting: %s\nDefault value is used." % error
            except AttributeError:
                pass
        return _value

    def _set_value(self, key, value):
        _type = _PREFS_TYPES[key]

        if _type == gconf.VALUE_STRING:
            try:
                if self._client.key_is_writable(key):
                    self._client.set_string(key, value)
            except glib.GError, error:
                print "Error while setting gconf value: %s\nError can be safely ignored: setting is not being changed." % error



#   Simple Backup - SFTP fuse plugin 
#
#   Copyright (c)2010: Jean-Peer Lorenz <peer.loz@gmx.net>
#   Copyright (c)2007-2008: Ouattara Oumar Aziz <wattazoum@gmail.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from sbackup.fuse_plugins import pluginFAM
import subprocess
import re
import os
from gettext import gettext as _
from tempfile import mkstemp
from sbackup.util.exceptions import FuseFAMException
from sbackup.util import local_file_utils


sftp_re = "^sftp://"
sftpUrlRegex = sftp_re + "(([^:]+):([^@]+)@)?" + "([^/^:^@]+?)" + "(:([0-9]+))?" + "/(.*)"

class sftpFuseFAM (pluginFAM)  :
    """
    The fuseFAM plugin for sftp
    @requires: curlftpfs
    """
    def __init__(self):
        pluginFAM.__init__(self)

    def match_scheme(self, remoteSource):
        _res = False
        _search_res = re.compile(sftp_re).search(remoteSource)
        if _search_res is not None:
            _res = True
        return _res

    def match_scheme_full(self, remoteSource):
        """This method checks for the scheme (the protocol) of the given
        remote source, i.e. it should not check the validity of the URL.
        This behavior is necessary since otherwise no plugin is found
        to handle a FTP address if this address is not valid. The user
        would be confused. The validity of a given FTP address must be
        checked separate.
        
        @todo: The plugins ssh, sftp do not behave if the suggested way! Fix them!
        """
        _res = False
        _search_res = re.compile(sftpUrlRegex).search(remoteSource)
        if _search_res is not None:
            _res = True
        return _res

    def mount(self, source, mountbase):
        """Mount the source into the mountbase dir . This method should
        create a mount point to mount the source. The name of the mount
        point should be very expressive so that we avoid collision with
        other mount points
        
        @param source: The remote path
        @param mountbase: The mount points base dir
        @return: The mount point complete path
        """
        #make the mount point
        spliturl = SplittedURL(source)
        mountpoint = self.__get_mount_dir(mountbase, spliturl)

        if not os.path.exists(mountpoint) :
            os.mkdir(mountpoint)

        #If the path is already mounted No need to retry
        if not self.checkifmounted(source, mountbase) :
            # Create output log file
            outptr, outFile = mkstemp(prefix = "sftpFuseFAMmount_output_")
            # Create error log file
            errptr, errFile = mkstemp(prefix = "sftpFuseFAMmount_error_")

            # the option 'allow_root' is necessary to grant access
            # if the script is invoked as superuser
            curl_cmd = ["curlftpfs", "-o", "direct_io"]

            if spliturl.user and spliturl.password:
                curl_cmd.append("-o")
                opts = "user=%s:%s" % (spliturl.user, spliturl.password)
                curl_cmd.append(opts)
            if os.getuid() == 0:
                curl_cmd.append("-o")
                curl_cmd.append("allow_root")

            server = spliturl.server
            if spliturl.port:
                server += ":" + spliturl.port

            curl_cmd.append(server)

            curl_cmd.append(mountpoint)

            # Call the subprocess using convenience method
            try:
                retval = subprocess.call(curl_cmd, 0, None, None, outptr, errptr)
            except OSError, _exc:
                os.rmdir(mountpoint)
                raise FuseFAMException(_("Couldn't found external application 'curlftpfs' needed for handling of sftp sites: %s") % _exc)

            # Close log handles
            os.close(errptr)
            os.close(outptr)
            outStr, errStr = local_file_utils.readfile(outFile), local_file_utils.readfile(errFile)
            local_file_utils.delete(outFile)
            local_file_utils.delete(errFile)
            if retval != 0:
                os.rmdir(mountpoint)
                raise FuseFAMException(_("Couldn't mount '%(server)s' into '%(mountpoint)s' : %(error)s") % {'server' : spliturl.server ,
                                                    'mountpoint': mountpoint,
                                                    'error':errStr})
        else:
            pass    # it is already mounted, do nothing

        remote_site = "sftp://" + spliturl.server
        if spliturl.port:
            remote_site += ":" + spliturl.port
        return (remote_site, mountpoint, spliturl.pathinside)

    def getdoc(self):
        """Returns a short documentation of this plugin.
        """
        doc = _("SFTP schema is like : sftp://user:pass@server/anything")
        return doc

    def checkifmounted(self, source, mountbase):
        """Checks if the given source is already mounted in given directory.
        
        @return: True if mounted, False if not
        """
        spliturl = SplittedURL(source)
        mountpoint = self.__get_mount_dir(mountbase, spliturl)
        return os.path.ismount(mountpoint)

    def __defineMountDirName(self, spliturl):
        """
        Helper method that builds the name of the mount directory.
        """
        dirname = "sftp_"
        if spliturl.user :
            dirname += spliturl.user + "@"
        dirname += spliturl.server
        if spliturl.port:
            dirname += "_%s" % spliturl.port
        return dirname

    def __get_mount_dir(self, mountbase, spliturl):
        """Helper method that builds the full path to the mountpoint.
        """
        mountpoint = os.path.join(mountbase, self.__defineMountDirName(spliturl))
        return mountpoint

class SplittedURL:
    """This will match the RE and give us a group like
        ('sftp://', 'test:pass@', 'wattazoum-vm.ft.nn', 'ddd/kjlh/klhkl/vvvv')
        
        @param remote: the remote site address
        @type remote: String
        
        @return: the address split into protocol, user, password, server
                 and path on server
        @rtype:  Tuple of Strings         
        """

    def __init__(self, url):

        exp = re.compile(sftpUrlRegex)
        match = exp.search(url)
        if match is None:
            raise FuseFAMException(_("Error matching the schema 'sftp://user:pass@server/anything' with '%s' (The '/' after server is mandatory)") % url)

        self.user = match.group(2)
        self.password = match.group(3)
        self.server = match.group(4)
        self.port = match.group(6)
        self.pathinside = match.group(7)



#   Simple Backup - definition of common constants related to SBackup
#
#   Copyright (c)2010: Jean-Peer Lorenz <peer.loz@gmx.net>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

#TODO: Add module dbus_constants?

# values in milliseconds if not stated otherwise in variable name


import signal


ONE_SECOND = 1
ONE_SECOND_IN_MILLISEC = 1000
SECONDS_IN_HOUR = 3600
SECONDS_IN_DAY = 24 * SECONDS_IN_HOUR
DAYS_IN_YEAR = 365
DAYS_IN_QUARTER = 91.25
DAYS_IN_MONTH = 30.5
DAYS_IN_WEEK = 7


TIMEOUT_RETRY_TARGET_CHECK_SECONDS = 300
INTERVAL_RETRY_TARGET_CHECK_SECONDS = 5

INDICATOR_LAUNCH_PAUSE_SECONDS = 3
TIMEOUT_INDICATOR_QUIT_SECONDS = 180 # timeout before quitting the indicator app

DBUS_CHECK_INTERVAL_SECONDS = 50
DBUS_RECONNECT_INTERVAL_SECONDS = 50
AUTOEXIT_TIMEOUT_SECONDS = 300
AUTOEXIT_CHECK_INTERVAL_SECONDS = 300

INTERVAL_UNREGISTER_DEAD_CONN = 600000  # milliseconds
DBUS_SERVICE_QUIT_PAUSE = 500


# settings
RETRY_UNKNOWN = -1
RETRY_TRUE = 1
RETRY_FALSE = 0
PID_UNKNOWN = -1
SPACE_REQUIRED_UNKNOWN = -1
FREE_SPACE_UNKNOWN = -1
SIZE_FILESYSTEM_UNKNOWN = -1
TARGET_UNKNOWN = ""
PROFILE_UNKNOWN = ""

# TAR
TAR_BLOCKSIZE = 512     # block size in Bytes, fixed in Tar, keep synced in multipleTarScript
TAR_BLOCKING_FACTOR = 20
TAR_RECORDSIZE = TAR_BLOCKING_FACTOR * TAR_BLOCKSIZE
TAR_VOLUME_SIZE_UNIT_IN_BYTES = 1024



##TODO: Keep copy of this used in `sbackup-terminate` up-to-date!
BACKUP_CANCEL_SIG = signal.SIGUSR1


# Name definitions
DBUSSERVICE_FILE = "sbackup-dbusservice"
INDICATORAPP_FILE = "sbackup-indicator"
INDICATORAPP_NAME = "Simple Backup Indicator Application"

DBUS_NOTIFIER_NAME = "Simple Backup DBus Notifier"

TERMINATE_FILE = "sbackup-terminate"
BACKUP_COMMAND = "sbackup"
BACKUP_PROCESS_NAME = "Simple Backup Process"


# icon filenames
DEFAULT_ICON_FILENAME = "sbackup.png"
NOTIFICATION_ICON_FILENAME = "sbackup32x32.png"
CONFIG_ICON_FILENAME = "sbackup-conf.png"
RESTORE_ICON_FILENAME = "sbackup-restore.png"

# icon names (must be registered before using)
INDICATOR_ATTENTION_ICON = "sbackup-attention"
INDICATOR_ACTIVE_ICON = "sbackup-panel"
INDICATOR_SUCCESS_ICON = "sbackup-success"


# used in module `util`
RSRC_FILE = "resources"
LOCALE_DIR = "locale"
LOCALE_DOMAIN = "sbackup"

NOTIFICATION_DOMAIN = "sbackup"


# dbus constants - keep copies of these up-to-date in `sbackup-progress`
#                  and `org.sbackupteam.SimpleBackup.conf` and `Makefile`
#                  configuration file name = dbus service!
DBUS_SERVICE = "org.sbackupteam.SimpleBackup"
DBUS_OBJ_PATH = "/SBackupProcess"
DBUS_CONNECTION_OBJ_PATH = "/SBackupConnection"
DBUS_INTERFACE = "org.sbackupteam.SimpleBackup.SBackupDbusInterface"
DBUS_EXCEPTION = "org.sbackupteam.SBackupDbusException"



# Global path definitions
LOCKFILE_BACKUP_FULL_PATH = "/var/lock/sbackup/sbackup.lock"
LOCKFILE_INDICATOR_FULL_PATH = "/var/lock/sbackup/sbackup-indicator.lock"


# Exit codes
EXCODE_SUCCESS = 0
EXCODE_GENERAL_ERROR = 1
EXCODE_INSTANCE_ALREADY_RUNNING = 3
EXCODE_KEYBOARD_INTERRUPT = 39
EXCODE_MAIL_ERROR = 38
EXCODE_BACKUP_ERROR = 38



#   Simple Backup - collection of files and files metadata
#
#   Copyright (c)2010: Jean-Peer Lorenz <peer.loz@gmx.net>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#


from gettext import gettext as _
import os
import re
import types

from sbackup import util

from sbackup.util import local_file_utils
from sbackup.util import exceptions
from sbackup.util import log


#TODO: make this module independent from archive backend 


class FileCollectorStats(object):
    """Provides statistical information about files collected by `FileCollector` objects.
    These information encompass:
    * size of files being backuped
    * number of files being backuped.
    """

    def __init__(self, followlinks = False):
        self.__followlinks = False
        self.set_followlinks(followlinks)
        # uncompressed size of snapshot (cummulative)
        self.__size_in_bytes = 0L

        self.__ndirs = 0L
        self.__nfiles = 0L
        self.__nsymlinks = 0L

        # for incremental counting (in case of full snapshot only `__nfile_incl` is used)
        self.__nfiles_incl = 0L
        self.__nfiles_skip = 0L
        self.__nfiles_new = 0L

        self.__nexcl_forced = 0L
        self.__nexcl_config = 0L

    def set_followlinks(self, followlinks):
        if not isinstance(followlinks, types.BooleanType):
            raise TypeError("Expected parameter of boolean type. "\
                            "Got %s instead." % type(followlinks))
        self.__followlinks = followlinks

    def get_size_payload(self):
        """Returns the cumulated size of files being backuped in bytes.
        Additional overhead due to storage etc. is not considered.
        """
        return self.__size_in_bytes

    def get_size_overhead(self, size_per_item):
        """Returns the cumulated size of overhead produced by files being backuped.
        The overhead is calculated based on the number of files and their overhead
        per file given as parameter.
        """
        if not isinstance(size_per_item, types.IntType):
            raise TypeError("Expected parameter of integer type. "\
                            "Got %s instead." % type(size_per_item))
        _overhead = (self.__ndirs + self.__nfiles_incl) * size_per_item
        if not self.__followlinks:
            _overhead += self.__nsymlinks * size_per_item
        return _overhead

    def get_count_files_total(self):
        """Returns the number of files.
        """
        return self.__nfiles

    def get_count_files_incl(self):
        """Returns the number of files.
        """
        return self.__nfiles_incl

    def get_count_files_skip(self):
        """Returns the number of files.
        """
        return self.__nfiles_skip

    def get_count_files_new(self):
        """Returns the number of new included files.
        """
        return self.__nfiles_new

    def get_count_dirs(self):
        """Returns the number of files.
        """
        return self.__ndirs

    def get_count_symlinks(self):
        """Returns the number of files.
        """
        return self.__nsymlinks

    def get_count_items_excl_forced(self):
        """Returns the number of files.
        """
        return self.__nexcl_forced

    def get_count_items_excl_config(self):
        """Returns the number of files.
        """
        return self.__nexcl_config

    def clear(self):
        """Clears collected data.
        """
        self.__size_in_bytes = 0L
        self.__ndirs = 0L
        self.__nfiles = 0L
        self.__nsymlinks = 0L
        self.__nfiles_incl = 0L
        self.__nfiles_skip = 0L
        self.__nexcl_forced = 0L
        self.__nexcl_config = 0L
        self.__nfiles_new = 0L

    def add_size(self, value):
        """The given value is added to the cumulated file size.
        """
        self.__size_in_bytes += value

    def count_file(self):
        """The file counter is increased by 1.
        """
        self.__nfiles += 1

    def count_incl_file(self):
        self.__nfiles_incl += 1

    def count_new_file(self):
        self.__nfiles_new += 1

    def count_skip_file(self):
        self.__nfiles_skip += 1

    def count_dir(self):
        """The file counter is increased by 1.
        """
        self.__ndirs += 1

    def count_symlink(self):
        self.__nsymlinks += 1

    def count_excl_forced(self):
        self.__nexcl_forced += 1

    def count_excl_config(self):
        self.__nexcl_config += 1


class FileCollectorParentSnapshotFacade(object):
    """Class that provides simplified access to attributes of the parent
    snapshot (the base). The class is designed for use with `FileCollector`.
    """

    def __init__(self):
        self.__logger = log.LogFactory.getLogger()
        # snapshot file (snar) of current snapshot's parent (base snapshot)
        # only set if the current one is incremental
        self.__base_snar = None
        self.__base_backup_time = None
        self.__base_snardict = None

    def set_base_snar(self, basesnar):
        """Sets the snapshot file (snar) of the base (parent) snapshot.
        
        @note: This implies that the current snapshot is incremental.
        
        """
#        if not isinstance(basesnar, tar.SnapshotFile):
#            raise TypeError("Expected parameter of type 'SnapshotFile'. "\
#                            "Got %s instead." % type(basesnar))
        self.__base_snar = basesnar
        self.__set_base_backup_time(basesnar.get_time_of_backup())
        self.__set_base_snardict(basesnar.get_dict_format2())

    def __set_base_backup_time(self, backup_time):
        """Sets the time the parent backup was created. The time is
        measured in seconds since beginning of the epoch (unix style).
        """
        if not isinstance(backup_time, types.FloatType):
            raise TypeError("Expected parameter of floar type. "\
                            "Got %s instead." % type(backup_time))
        self.__logger.debug("Backup time of parent snapshot: %s" % backup_time)
        self.__base_backup_time = backup_time

    def __set_base_snardict(self, snardict):
        """Sets the dictonary containing the parent snapshot file (snar file).
        """
        if not isinstance(snardict, types.DictionaryType):
            raise TypeError("Expected parameter of dictionary type. "\
                            "Got %s instead." % type(snardict))
        self.__base_snardict = snardict

    def get_base_snardict(self):
        """Returns the dictonary containing the parent snapshot file (snar file).
        """
        return self.__base_snardict

    def get_base_backup_time(self):
        """Returns the time the parent backup was created.
        """
        return self.__base_backup_time


class FileCollector(object):
    """Responsible for the process of collecting files that are being backuped.
    The collecting process comprises of:
    
    * check the files are readable/accessable
    * apply exclusion rules (Regex) defined by user to the list of files
    * calculate the required space for the backup
    * prepare include and exclude file lists used by the backup process.
    """

    def __init__(self, snp, configuration):
        self.__logger = log.LogFactory.getLogger()

        # current snapshot (the new one)
        self.__snapshot = None
        # flag whether the current snapshot is full or incremental
        self.__isfull = True

        self.__parent = FileCollectorParentSnapshotFacade()
        self.__collect_stats = FileCollectorStats()
        self.__configuration = None

        # stats of last processed file
        self.__stat_func = None
        self.__fstats = None
        self.__fislink = None
        self.__fisdir = None
        # list of Regular Expressions defining exclusion rules
        self.__excl_regex = []
# TODO: put list of compiled regex into `Snapshot` (i.e. compile them when setting the excludes). 

        self.set_snapshot(snp)
        self.set_configuration(configuration)

    def set_snapshot(self, snp):
        """Sets the given snapshot as the currently processed snapshot.
        """
#        if not isinstance(snp, snapshot.Snapshot):
#            raise TypeError("Expected parameter of type 'Snapshot'. "\
#                            "Got %s instead." % type(snp))
        self.__snapshot = snp
        self.__set_isfull(isfull = snp.isfull())
        self.__collect_stats.set_followlinks(followlinks = snp.isFollowLinks())

    def set_configuration(self, configuration):
        """Sets the given object of type `FileCollectorConfigFacade`.
        """
        if not isinstance(configuration, FileCollectorConfigFacade):
            raise TypeError("Expected parameter of type 'FileCollectorConfigFacade'. "\
                            "Got %s instead." % type(configuration))
        self.__configuration = configuration

    def set_parent_snapshot(self, parent):
        """Sets the snapshot file (snar) of the base (parent) snapshot.
        
        @note: This implies that the current snapshot is incremental.
        
        """
#        if not isinstance(parent, tar.SnapshotFile):
#            raise TypeError("Expected parameter of type 'SnapshotFile'. "\
#                            "Got %s instead." % type(parent))
        self.__set_isfull(isfull = False)
        self.__parent.set_base_snar(parent)

    def get_stats(self):
        """Returns the collector stats object.
        """
        return self.__collect_stats

    def __prepare_collecting(self):
        """The actual process of collecting is prepared (i.e. stats are cleared etc.).
        
        @note: Depending on setting `Follow links` are functions for testing file existance and
               retrieval of file stats selected here.
        """
        if self.__snapshot.isFollowLinks():
            self.__stat_func = os.stat
        else:
            self.__stat_func = os.lstat
        self.__collect_stats.clear()

    def __set_isfull(self, isfull):
        """Sets attribute `__isfull` to the given boolean value.
        
        Attribute `__isfull` is introduced because of performance concerns
        since the snapshot derives this information from its name on every
        request again.  
        """
        if not isinstance(isfull, types.BooleanType):
            raise TypeError("Expected parameter of boolean type. "\
                            "Got %s instead." % type(isfull))
        self.__isfull = isfull
        # test of post-condition: isfull must be equal to the value in the snapshot
        if self.__isfull != self.__snapshot.isfull():
            raise AssertionError("Values of attribute 'isfull' are "\
                                 "inconsistent. Found value in snapshot: %s."\
                                 % self.__snapshot.isfull())

    def __is_not_accessable(self, path):
        """Tests whether the given `path` can be accessed (i.e. exists and is readable).
        """
        # get the stats, If not possible, the file has to be exclude, return True
        try:
            self.__fstats = self.__stat_func(path)
            self.__fisdir = local_file_utils.is_dir(path)
            self.__fislink = local_file_utils.is_link(path)
        except Exception, _exc:    #IGNORE:W0703
            self.__logger.warning(_("File '%(file)s' is not accessable with error '%(error)s'.")\
                                    % {'file' : path, 'error' : str(_exc)})
            self.__fstats = None
            self.__fisdir = None
            self.__fislink = None
            return True

        # refuse a file if we don't have read access
        # The open() statement may hang indefinitely (LP Bug 184713)
        # taken from: http://docs.python.org/library/signal.html#example
        _res = False
        try:
            util.set_timeout_alarm(timeout = 5)
            fdscr = os.open(path, os.R_OK)
            os.close(fdscr)
        except exceptions.TimeoutError, _exc:
            self.__logger.warning(_("File '%(file)s' cannot be opened for read access. Operation timed out.")\
                                    % {'file' : path})
            _res = True
        except Exception, _exc:    #IGNORE:W0703
            self.__logger.warning(_("File '%(file)s' cannot be opened for read access with error '%(error)s'.")\
                                    % {'file': path, 'error' : str(_exc)})
            _res = True

        util.set_timeout_alarm(timeout = 0)
            # when file does not exist, the `open` fails and variable `fdscr` is not defined here
#            try:
#                os.close(fdscr)
#            except OSError:
#                pass    
        return _res

    def __is_circular_symlink(self, path):
        if self.__fislink:
            if self.__snapshot.isFollowLinks():
                ln_target = local_file_utils.get_link_abs(path)
                if path.startswith(ln_target):
                    self.__logger.info(_("Symbolic link '%(path)s' -> '%(ln_target)s' is a circular symlink.")\
                                % {'path' : path, 'ln_target' : ln_target})
                    return True
        #test passed
        return False

    def _is_excluded_by_name(self, path):
        """Decides whether or not a path has to be excluded by
        its name using the lists of defined
        * snapshot destination
        * regular expressions
        * blacklisted names.
                
        @return: True if the file has to be excluded, false if not
        """
        if path == self.__configuration.get_target_dir():
            self.__logger.info(_("File '%(file)s' is backup's target directory.") % {'file' : path})
            return True
        
        # if the file is in exclude list, return true
        if self.__snapshot.is_path_in_excl_filelist(path):
            self.__logger.info(_("Path '%(file)s' defined in excludes list.") % {'file' : path})
            return True

        # if the file matches an exclude regexp, return true
# TODO: Regex are applied to the full path. Add a choice to apply Regex only to files, directories etc.
        for _regex in self.__excl_regex:
            _regex_res = _regex.search(path)
            if _regex_res is not None:
                self.__logger.info(_("File '%(file)s' matches regular expression '%(regex)s'.")\
                                    % {'file' : path, 'regex' : str(_regex.pattern)})
                return True
        #all tests passed
        return False

    def _is_excluded_by_size(self, path):
        """Decides whether or not a file is to be excluded by the configuration.
        It is not decided for the incremental exclusion.

        Currently, following configuration options are tested:
        * file size
                
        @return: True if the file has to be excluded, false if not
        """
        #if the file is too big
        if self.__configuration.is_maxsize_enable():
            if self.__fstats.st_size > self.__configuration.get_maxsize_limit():
                self.__logger.info(_("File '%(file)s' exceeds maximum file size ( %(filesize)s > %(maxsize)s).")\
                                    % {'file' : path, 'filesize' : str(self.__fstats.st_size),
                                       'maxsize' : str(self.__configuration.get_maxsize_limit())})
                return True
        #all tests passed
        return False

    def _is_excluded_by_force(self, path):
        """Private interface method which checks for forced exclusion of given `path` by
        calling the according test methods in turn. If this method returns True, the
        path *must* be excluded irrespectively it is explicitely included etc.

        @return: True if the file has to be excluded, false if not
        """
        if self.__is_not_accessable(path) is True:
            return True
        elif self.__is_circular_symlink(path) is True:
            return True
        return False

    def _check_for_excludes(self, path): #, force_exclusion=False):
        """Checks given `path` for exclusion and adds it to the `ExcludeFlist` if
        required. Sub-directories are only entered in the case the `path` is not
        excluded.
        
        @param path: The path being checked for exclusion
        
        @note: Links are always backuped; TAR follows links (i.e. dereferences them = stores the actual
               content) only if option `followlinks` is set. A link targeting a directory yields
               'islink=True' and 'isdir=True'. 
        """
        _excluded = False
        _stop_checking = False

        if self._is_excluded_by_name(path):
            if not self.__snapshot.is_subpath_in_incl_filelist(path):
                # add to exclude list, if not explicitly included; since paths can be nested,
                # it is checked for sub-paths instead of full paths
                self.__snapshot.addToExcludeFlist(path)
                self.__collect_stats.count_excl_config()
                _excluded = True
            
        elif self._is_excluded_by_force(path):
            # force exclusion e.g. path is defined in includes list but does not exist/is not accessable
            self.__snapshot.addToExcludeFlist(path)
            self.__collect_stats.count_excl_forced()
            _excluded = True

        elif self._is_excluded_by_size(path):
            if not self.__snapshot.is_subpath_in_incl_filelist(path):
                # add to exclude list, if not explicitly included; since paths can be nested,
                # it is checked for sub-paths instead of full paths
                self.__snapshot.addToExcludeFlist(path)
                self.__collect_stats.count_excl_config()
                _excluded = True

        if not _excluded:
            # path was not excluded, so do further tests (stats, enter dir...)            
            if self.__fislink:
                self.__logger.debug("Symbolic link found: '%(path)s' -> '%(ln_target)s'."\
                                % {'path' : path, 'ln_target' : local_file_utils.get_link(path)})
                self.__collect_stats.count_symlink()
                if not self.__snapshot.isFollowLinks():
                    # if `followlinks` is *disabled*, just count the link and finish
                    _stop_checking = True

            if self.__fisdir:
                if _stop_checking:    # i.e. `followlinks` is not enabled
                    self.__collect_stats.count_file()
                    self.__cumulate_size(path)
                else:
                    # if it's a directory, enter inside
                    try:
                        for _dir_item in local_file_utils.listdir(path) :
                            _dir_item = local_file_utils.joinpath(path, _dir_item)
                            self._check_for_excludes(path = _dir_item)
                        self.__collect_stats.count_dir()    # the directory `path`
                    except OSError, _exc:
                        self.__logger.warning(_("Error while checking directory '%(dir)s': %(error)s.")\
                                              % {'dir' : path, 'error' : str(_exc)})
                        self.__snapshot.addToExcludeFlist(path)    # problems with `path` -> exclude it
                        self.__collect_stats.count_excl_forced()
            else:
                # it's a file (may also a link target in case of enabled `followlinks` option)
                self.__collect_stats.count_file()
                self.__cumulate_size(path)

    def __cumulate_size(self, path):
        """
        
        Files not contained in SNAR file are backuped in any case!
        (e.g. a directory was added to the includes)
        """
#        self.__logger.debug("%s" % path)
        _incl_file = False
        if self.__isfull:        # full snapshots do not have a base snar file
            _incl_file = True
        else:
            # we don't look at the access time since this was even modified during the last backup 
            ftime = max(self.__fstats.st_mtime,
                        self.__fstats.st_ctime)
            if path in self.__parent.get_base_snardict():
#                self.__logger.debug("%s: is in snapshot file." % path)
                if ftime > self.__parent.get_base_backup_time():
                    self.__logger.debug("Delta=%s - %s: %s > %s" % ((ftime - self.__parent.get_base_backup_time()),
                                                                     path, ftime,
                                                                     self.__parent.get_base_backup_time()))
                    _incl_file = True

            else:
                self.__logger.debug("%s: No yet included - included." % path)
                self.__collect_stats.count_new_file()
                _incl_file = True

        if _incl_file:
            self.__collect_stats.count_incl_file()
            self.__collect_stats.add_size(self.__fstats.st_size)
        else:
            self.__collect_stats.count_skip_file()

    def __compile_excl_regex(self):
        """Prepares (i.e. compiles) Regular Expressions used for excluding files from flist.
        """
        self.__logger.debug("Prepare Regular Expressions used for file exclusion.")
        _rexclude = []
        _snp_excl = self.__snapshot.getExcludes()
        if _snp_excl:
            for _regex in _snp_excl:
                if util.is_empty_regexp(_regex):
                    self.__logger.warning(_("Empty regular expression found. Skipped."))
                else:
                    if util.is_valid_regexp(_regex):
                        _regex_c = re.compile(_regex)
                        _rexclude.append(_regex_c)
                    else:
                        self.__logger.warning(_("Invalid regular expression ('%s') found. Skipped.") % _regex)
        self.__excl_regex = _rexclude

    def __prepare_explicit_flists(self):
        """Paths (i.e. directories and files) defined in the configuration are added
        to the corresponding include/exclude file lists. These lists represent the
        explicitely defined includes and excludes. Includes are stronger than excludes!
        
        The config GUI avoids the definition of a file as included and excluded at the same time.
        The `ConfigManager` returns the last defined entry in the case of multiple definitions
        in section `dirconfig`.        
        """
        _config = self.__configuration
        if _config is None:
            raise ValueError("No configuration object set.")

        # set the list to backup (includes) and to exclude
        self.__logger.debug("Preparation of include and exclude lists.")
        if _config.has_dirconfig_entries() is False:
            self.__logger.warning(_("No directories to backup defined."))
        else:
            for _dir, _incl in _config.get_dirconfig_local():
                if _incl == 1:
                    self.__snapshot.addToIncludeFlist(_dir)
                elif _incl == 0:
                    self.__snapshot.addToExcludeFlist(_dir)

        # add the default excluded ones
# TODO: why add an empty string?
# TODO: put the default excludes into `ConfigManager`.
# TODO: Remove this hack and replace it by proper shell pattern expansion (glob).
        _excludes = ["", "/dev/", "/proc/", "/sys/", "/tmp/"]
        if _config.get_target_dir() is not None:
            _excludes.append(_config.get_target_dir().rstrip(os.sep) + os.sep)
        for _excl in _excludes:
            self.__snapshot.disable_path_in_incl_filelist(_excl)

# TODO: Even better: Remove this inconsistency and don't use TAR's shell patterns. Prefer pure Regex exclusion.
        # LP #638072: no wildcards in excludes list;
        #             excluded directory yields in excluding contained files
#FIXME: trailing slashes are not retained when writing excludes file! Since exclude items
#       are literally compared this can exclude more dirs/files than intended!
        for _excl in _excludes:
            self.__snapshot.addToExcludeFlist(_excl)

        # sanity check of the lists
        self.__snapshot.check_and_clean_flists()

    def collect_files(self):
        """Collects information about files that are included resp. excluded in
        the backup.
        """
        self.__prepare_collecting()
        self.__compile_excl_regex()
        self.__prepare_explicit_flists()
        util.enable_timeout_alarm()

        # We have now every thing we need , the rexclude, excludelist, includelist and already stored 
        self.__logger.debug("Creation of the complete exclude list.")

        # walk recursively into paths defined as includes (therefore don't call nested paths)
        for _incl in self.__snapshot.get_eff_incl_filelst_not_nested():
            _incl = local_file_utils.normpath(_incl)
            self._check_for_excludes(_incl)


class FileCollectorConfigFacade(object):
    """Provides a simplified and unified read-only access to configuration options required
    for file collection. Purpose is to de-couple class `ConfigManager` and `FileCollector`
    since only very little configuration settings are required in order to collect the files.
    """
    def __init__(self, configuration, eff_local_targetdir):
        self.__configuration = None
        self.__eff_local_targetdir = eff_local_targetdir     # needed to exclude paths == target

        self.__maxsize_enabled = False
        self.__maxsize = 0

        self.__dirconfig = None
        self.__dirconfig_set = False

        self.set_configuration(configuration)

    def set_configuration(self, configuration):
        """Sets the given `configuration`. Corresponding attributes are
        set from this ConfigManager instance.
        """
#        if not isinstance(configuration, ConfigManager):
#            raise TypeError("Expected parameter of type 'ConfigManager'. "\
#                            "Got %s instead." % type(configuration))
        self.__configuration = configuration
        self.__set_maxsize_limit_from_config()
        self.__set_dirconfig_from_config()

    def __set_maxsize_limit_from_config(self):
        if self.__configuration is None:
            raise ValueError("No configuration set.")
        self.__maxsize_enabled = self.__configuration.has_maxsize_limit()
        self.__maxsize = self.__configuration.get_maxsize_limit()

    def __set_dirconfig_from_config(self):
        if self.__configuration is None:
            raise ValueError("No configuration set.")
        self.__dirconfig = self.__configuration.get_dirconfig_local()
        if self.__dirconfig is None:
            self.__dirconfig_set = False
        else:
            self.__dirconfig_set = True

    def is_maxsize_enable(self):
        return self.__maxsize_enabled

    def get_maxsize_limit(self):
        return self.__maxsize

    def get_target_dir(self):
        return self.__eff_local_targetdir

    def get_dirconfig_local(self):
        """Returns the directory configuration stored in a list of pairs (name, value).
        
        @rtype: List
        """
        return self.__dirconfig

    def has_dirconfig_entries(self):
        """Returns True if the `dirconfig` has any entries. Purpose is improvement of
        performance when checking for entries in the `dirconfig` and hiding implementation
        details about format/storage of the list.
        """
        return self.__dirconfig_set
